"""The processor base class."""

import gc
import json
import logging
import os
from collections.abc import Sequence
from pathlib import Path
from typing import Any, Literal, Protocol, Self, runtime_checkable

import matplotlib.pyplot as plt
import nexusformat.nexus as nx
import numpy as np

from . import signal_processing as sp
from ._data import DataProcessed, DataRaw

# Global constants
PROGRAM_NAME = "pymagnetos"  # program name, saved in the NeXus file
RESULTS_NAME = "results"  # name of the main NeXus group
SERIE_NAME = "serie"  # name of the data series group in the NeXus file
NEXUS_EXTENSION = ("nx5", "nxs", "h5", "hdf5")  # NeXus file extension

logger = logging.getLogger(__name__)


@runtime_checkable
class Config(Protocol):
    def loads(): ...


class BaseProcessor:
    """
    Load, analyze and store data acquired during pulsed magnetic field experiment.

    This is a base class meant to be subclassed to create specialized processor objects.

    It provides methods shared between experiments :
    - Load a configuration file (via the `Config` object),
    - Load binary and text data,
    - Store the data in-memory as NeXus objects (via the `Data` objects),
    - Compute a magnetic field over time,
    - Save the results as a NeXus file.

    It offers a convenient way to get and set data in a NeXus hierarchy (raw and
    processed data, sorted in different series) and can dump it in a NeXus-compliant
    HDF5 file along with all the metadata required to recreate a Processor object from
    such a file. The file can be optionnaly consolidated by specifying additional NeXus
    fields in the configuration file.

    The `_config_cls` attribute should be a Config class (the type, not an instance) and
    be set by subclasses.
    """

    _config_cls: type[Config]
    measurements: list[str]

    def __init__(self, file_path: Path | str | None = None, **kwargs) -> None:
        """
        Load, analyze and store data acquired during pulsed magnetic field experiment.

        Parameters
        ----------
        file_path : Path, str or None, optional
            Full path to the configuration file or a previously saved NeXus file. Can be
            None (default) to instantiate an empty object.
        **kwargs : passed to the `load_file()` method.
        """
        # Prepare internal variables in case it was not set before
        if not hasattr(self, "_results_name"):
            self._results_name = RESULTS_NAME
        if not hasattr(self, "_serie_name"):
            self._serie_name = SERIE_NAME
        if not hasattr(self, "_program_name"):
            self._program_name = PROGRAM_NAME
        if not hasattr(self, "_config_cls"):
            raise AttributeError(
                "No Config class defined, subclasses must set the '_config_cls' "
                "attribute"
            )

        # Load configuration or NeXus and trigger further initializations
        self.load_file(file_path, **kwargs)

    @property
    def expid(self) -> str:
        """Experiment ID, linked to the Config object."""
        return self.cfg.expid

    @expid.setter
    def expid(self, value: str):
        """
        Setter for `expid`.

        The `expid` attribute of the configuration is updated. The filenames are
        rebuilt.
        """
        self.cfg.expid = value
        self.cfg.build_filenames()
        self._reinit()
        logger.info(f"Experiment ID set to '{self.expid}'")

    @property
    def data_directory(self) -> Path:
        """Data directory, linked to the Config object."""
        return self.cfg.data_directory

    @data_directory.setter
    def data_directory(self, value: str | Path):
        """
        Setter for `data_directory`.

        The `data_directory` attribute of the configuration is updated. The filenames
        are rebuilt.
        """
        self.cfg.data_directory = Path(value)
        self.cfg.build_filenames()
        self._reinit()
        logger.info(f"Data directory set to '{self.data_directory}'")

    @property
    def idx_serie(self) -> Any:
        """Current analysis serie index."""
        return -1

    @idx_serie.setter
    def idx_serie(self, value: Any):
        raise NotImplementedError("Subclasses must implement this property setter")

    @property
    def analysis_window(self) -> Sequence[float]:
        """Current analysis window."""
        return [0.0]

    @analysis_window.setter
    def analysis_window(self, value: Sequence[float]):
        raise NotImplementedError("Subclasses must implement this property setter")

    def __repr__(self) -> str:
        return f"{type(self)}\nDataset : {self.cfg.expid}"

    def __str__(self) -> str:
        return f"Dataset : {self.cfg.expid}"

    def _reinit(self) -> None:
        """Reinitialize the Processor, cleaning up the data."""
        self._init()
        gc.collect()

    def load_file(self, filepath: str | Path | None, **kwargs) -> None:
        """
        Load a configuration or NeXus file with the Config class.

        If loading a NeXus file, its extension should be "nx5", "nxs", "h5" or "hdf5" to
        be detected as such.

        Parameters
        ----------
        filepath : str, Path or None
            Full path to the configuration file or NeXus file.
        **kwargs : passed to the Config class.
        """
        if filepath is not None and str(filepath).endswith(NEXUS_EXTENSION):
            # Load the NeXus file instead
            self.is_config_file = False
            self.cfg = self._config_cls(**kwargs)  # initialize Config with defaults
            self._init()  # initialize default Processor
            self.load(filepath)  # load NeXus file
        else:
            # Load the configuration file
            self.is_config_file = True
            self.is_nexus_file = False
            self.cfg = self._config_cls(user_file=filepath, **kwargs)
            self.cfg.resolve_nexus(self._serie_name)
            self._init()  # initialize with Config

    def load_pickup(self) -> Self:
        """
        Load the pickup binary file, gathering metadata (if any) and the file.

        Pickup binary file might be different accross experiments, so subclasses must
        implement this method.
        """
        raise NotImplementedError("Subclasses must implement this method")

    def compute_field(self, method: str = "trapz") -> Self:
        """
        Compute magnetic field by integrating the pickup signal.

        The details are in the `_compute_field()` method.

        Parameters
        ----------
        method : str, optional
            Method to perform integration, passed to `sp.integrate_pickup()`. Default is
            "trapz" (which is only method currently supported).
        """
        # Checks
        if "pickup" not in self.data_raw:
            self.load_pickup()

        # Get parameters
        surface = self.cfg.parameters.pickup_surface

        # Compute
        self._compute_field(surface=surface, method=method)

        # Get up/down indices
        self.get_up_down_indices()

        return self

    def load_data(self) -> Self:
        """Subclass should implement this method to load the raw data."""
        raise NotImplementedError("Subclasses must implement this method")

    def preprocess(self) -> Self:
        """Subclass should implement this method to pre-process the data."""
        raise NotImplementedError("Subclasses must implement this method")

    def align_field(self, on_pickup: bool = False, **kwargs) -> Self:
        """
        Align magnetic field time serie on the experiment time vector.

        Optionally, also align the pickup signal. In that case, it will be stored in
        `data_processed`.

        **kwargs are passed to `numpy.interp()`.
        """
        if not self._check_field_computed():
            logger.warning(
                "Magnetic field not computed, cannot align on experiment time base"
            )
            return self

        # Collect data
        exptime = self.get_data_processed("time_exp")
        magtime = self.get_data_processed("magfield_time")
        magfield = self.get_data_processed("magfield")

        # Interpolate
        logger.info("Aligning magnetic field on experiment time base...")
        self.set_data_processed(
            "magfield", np.interp(exptime, magtime, magfield, **kwargs)
        )
        # Update the time vector
        self.set_data_processed("magfield_time", exptime)
        logger.info("Field aligned")

        # Also align the pickup signal if requested
        if on_pickup:
            logger.info("Aligning pickup signal on experiment time base...")
            pickup = self.get_data_raw("pickup")
            self.set_data_processed(
                "pickup", np.interp(exptime, magtime, pickup, **kwargs)
            )
            logger.info("Pickup signal aligned")

        # Get the new field up/down indices
        self.get_up_down_indices()

        return self

    def run_analysis(self) -> Self:
        """Subclasses should implement this method to perform the analysis."""
        raise NotImplementedError("Subclasses must implement this method")

    def save(
        self, filename: str | Path | None = None, consolidate: bool = True, **kwargs
    ) -> bool:
        """
        Save object to a NeXus[1] compliant HDF5 file.

        It is written using the `NXFile` object (that is an `h5py.File` subclass).
        If a filename is not specified, the `get_nexus_filename()` is used to get a path
        to the output file.

        The `Config` object and the `metadata` attribute are serialized and stored in
        the file to allow for reconstruction of the processing state.

        [1] https://www.nexusformat.org/

        Parameters
        ----------
        filename : str, Path or None
            Full path to the output file, that should't exist. If None (default), the
            filename is determined with the `get_nexus_filename()` method.
        consolidate : bool, optional
            Whether to consolidate the file to add attributes and groups from the
            configuration file. Default is True.
        **kwargs : passed to `nx.NXFile()`.

        Returns
        -------
        status : bool
            True if the file was saved successfully, False otherwise.
        """
        if not filename:
            filename = self.get_nexus_filename()
        if "mode" not in kwargs:
            kwargs["mode"] = "w-"

        logger.info("Saving NeXus file...")
        # Consolidation
        if consolidate:
            self.consolidate()
        self.resolve_nexus_links()

        # Specify configuration for the program
        self.data_processed["configuration"] = self.cfg.model_dump_json()
        self.data_processed["configuration"].attrs["format"] = "json"

        # Add metadata
        self.data_processed["metadata"] = json.dumps(self.metadata)
        self.data_processed["metadata"].attrs["format"] = "json"

        # Save
        try:
            with nx.NXFile(filename, **kwargs) as f:
                f.writefile(self.nxroot)
            # release
            self.release_nexus_file()
        except Exception as e:
            logger.error(f"Failed to save: {e}")
            return False

        logger.info(f"Saved at {filename}")
        return True

    def load(self, filename: str | Path | None) -> Self:
        """
        Load a previously created NeXus file.

        Note that only the first entry of the `NXroot` object will be loaded.

        Parameters
        ----------
        filename : str
            Full path to the file to load.

        Returns
        -------
        obj : BaseProcessor
            Initialized Processor object ready to carry on analysis.
        """
        # Load file
        logger.info(f"Loading NeXus file at {filename}...")
        try:
            with nx.NXFile(filename, mode="rw", recursive=True) as f:
                nxroot = f.readfile()
        except Exception as e:
            logger.error(f"\nFailed to load: {e}")
            return self

        # Get first entry
        nxentry = next(iter(nxroot.entries.values()))

        # Re-generate Config object
        if "configuration" in nxentry["processed"]["analysis"]:
            self.cfg = self._config_cls.loads(
                nxentry["processed"]["analysis"]["configuration"].nxdata
            )
        else:
            logger.warning(
                "No configuration found in the NeXus file, features will be limited"
            )

        # Set data
        # Processed data
        for key, value in nxentry["processed"]["analysis"].attrs.items():
            # attributes
            self.data_processed.attrs[key] = value
        for key, value in nxentry["processed"]["analysis"].items():
            # groups and fields
            self.data_processed[key] = value
        # Raw data
        for key, value in nxentry["raw"]["data"].attrs.items():
            # attributes
            self.data_raw.attrs[key] = value
        for key, value in nxentry["raw"]["data"].items():
            # groups and fields
            self.data_raw[key] = value

        # Re-build NX data structure
        name_entry = self.cfg.nexus.groups["root"]["name"]
        self.nxentry.nxname = name_entry
        self.nxroot.attrs["default"] = name_entry

        # Recover metadata
        if "metadata" in self.data_processed:
            self.metadata = json.loads(self.data_processed["metadata"].nxdata)
        else:
            logger.warning(
                "No metadata found in the NeXus file, features will be limited"
            )

        # Update flag
        self.is_nexus_file = True

        # Get up/down indices
        self.get_up_down_indices()

        logger.info("NeXus file loaded")

        return self

    @staticmethod
    def load_bin(
        filepath: str | Path,
        precision: int = 8,
        endian: Literal["<", ">"] = "<",
        **kwargs,
    ) -> np.ndarray:
        """
        Read a binary file with the given precision and endian.

        Simple wrapper around `numpy.fromfile()`.

        Parameters
        ----------
        filepath : str or Path
            Full path to the file to read.
        precision : int, optional
            Floating-point precision, by default 8.
        endian : {"<", ">"}, optional
            "<" for little endian, ">" for big endian, by default "<".
        **kwargs : passed to `numpy.fromfile()`.

        Returns
        -------
        data : np.ndarray
            Raw data from binary file.

        """
        with open(filepath, "rb") as fid:
            data = np.fromfile(fid, dtype=f"{endian}f{precision}", **kwargs)

        return data

    def create_fake_pickup(self) -> Self:
        """
        Create a fake pickup signal and store it in `data_raw`.

        Requires "time_exp" in `data_processed`.
        """
        self.set_data_raw(
            "pickup",
            self._make_fake_pickup(self.get_data_processed("time_exp").shape[0]),
        )
        self.set_data_raw("pickup_time", self.get_data_processed("time_exp"))

        return self

    def get_up_down_indices(self) -> Self:
        """Get increasing and decreasing field indices."""
        if self.get_data_processed("magfield", checkonly=True):
            self.inds_b_inc, self.inds_b_dec = sp.get_up_down_indices(
                self.get_data_processed("magfield")
            )

        return self

    def get_data_raw(self, name: str, checkonly: bool = False) -> Any:
        """
        Retrieve data from `data_raw[name]`.

        Note : the returned value is a *view*, if it is modified, the underlying data
        is modified as well.

        Parameters
        ----------
        name : str
            Dataset name.
        checkonly : bool, optional
            Perform a key existence check without returning the value. Default is False.
        """
        if checkonly:
            return name in self.data_raw
        else:
            return self.data_raw[name].nxdata

    def set_data_raw(self, name: str, value: Any) -> None:
        """
        Put data in `data_raw[name]`.

        Dataset will be written as an NXfield in `data_raw[name]`. If the dataset
        already exists, it is replaced (modifying its type and shape accordingly).

        If the dataset name is present in the 'nexus' section of the configuration, the
        attributes are written as well.

        Parameters
        ----------
        name : str
            Dataset name.
        value : Any
            Dataset.
        """
        if name in self.data_raw:
            self.data_raw[name].replace(value)
        else:
            self.data_raw[name] = value

        # Set dataset attributes
        if name in self.cfg.nexus.datasets:
            for k, v in self.cfg.nexus.datasets[name].items():
                self.data_raw[name].attrs[k] = v

    def get_data_processed(self, name: str, checkonly: bool = False) -> Any:
        """
        Get data from the 'results' group of the DataProcessed object.

        The data is pulled from `data_processed[results][name]`. If `checkonly` is True,
        only check if the dataset exists.

        Note : the returned value is a *view*, if it is modified, the underlying data
        is modified as well.

        Parameters
        ----------
        name : str
            Dataset name.
        checkonly : bool, optional
            Perform a key existence check without returning the value. Default is False.
        """
        where = f"{self._results_name}"
        return self._get_data(where, name, checkonly=checkonly)

    def set_data_processed(self, name: str, value: Any) -> None:
        """
        Set data in the 'results' group of the DataProcessed object.

        Dataset will be written as an NXfield in `data_processed[results][name]`. If the
        dataset already exists, it is replaced (modifying its type and shape
        accordingly).

        If the group name or the dataset name are present in the 'nexus' section of the
        configuration, the attributes are written as well.

        Parameters
        ----------
        name : str
            Dataset name.
        value : Any
            Dataset.
        """
        where = f"{self._results_name}"
        self._set_data(where, name, value)

    def set_map_in_data_processed(self, data: dict[str, np.ndarray]) -> None:
        """Set data in `data_processed`."""
        for key, value in data.items():
            self.set_data_processed(key, value)

    def remove_data_processed(self, name: str) -> None:
        """
        Remove dataset from the 'results' group of the DataProcessed object.

        The dataset stored at `data_processed[results][name]` is removed.

        Parameters
        ----------
        name : str
            Dataset name.
        """
        where = f"{self._results_name}"
        self._remove_data(where, name)

    def create_data_serie(self) -> None:
        """
        Check if the dictionnary for serie data exists, if not, create it.

        The group hosting the data for the current `idx_serie` is stored at :
        `data_processed[results_{serie_name}{idx_serie}]`.

        If 'results_{serie_name}' is present in the 'nexus' section of the
        configuration, the attributes are written as well.
        """
        baseloc = f"{self._results_name}_{self._serie_name}"
        loc = f"{baseloc}{self.idx_serie}"
        if loc not in self.data_processed:
            self.data_processed.create_serie(self.idx_serie)

            # Set NeXus attributes
            if baseloc in self.cfg.nexus.groups:
                for k, v in self.cfg.nexus.groups[baseloc].items():
                    self.data_processed[loc].attrs[k] = v

    def get_data_serie(self, name: str, checkonly: bool = False) -> Any:
        """
        Get data from the current 'results_serie' group of the DataProcessed object.

        The data is pulled from `data_processed[results_{serie_name}{idx_serie}][name]`.
        If `checkonly` is True, only check if the dataset exists.

        Note : the returned value is a *view*, if it is modified, the underlying data
        is modified as well.

        Parameters
        ----------
        name : str
            Dataset name.
        checkonly : bool, optional
            Perform a key existence check without returning the value. Default is False.
        """
        where = f"{self._results_name}_{self._serie_name}{self.idx_serie}"
        return self._get_data(where, name, checkonly=checkonly)

    def set_data_serie(self, name: str, value: Any) -> None:
        """
        Set data in the current 'results_serie' group of the DataProcessed object.

        Dataset will be written as an NXfield in
        `data_processed[results_{serie_name}{idx_serie}][name]`. If the dataset already
        exists, it is replaced (modifying its type and shape accordingly).

        If the group name or the dataset name are present in the 'nexus' section of the
        configuration, the attributes are written as well.

        Parameters
        ----------
        name : str
            Dataset name.
        value : Any
            Dataset.
        """
        where = f"{self._results_name}_{self._serie_name}{self.idx_serie}"
        self._set_data(where, name, value)

    def remove_date_serie(self, name: str) -> None:
        """
        Remove data in the current 'results_serie' group of the DataProcessed object.

        The dataset at `data_processed[results_{serie_name}{idx_serie}][name]` is
        removed.

        Parameters
        ----------
        name : str
            Dataset name.
        """
        where = f"{self._results_name}_{self._serie_name}{self.idx_serie}"
        self._remove_data(where, name)

    def set_attr_serie(self, name: str, value: Any) -> None:
        """
        Set attribute for the current 'results_serie' NXdata group in data_processed.

        The attribute will be written for the group at
        `data_processed[results_{serie_name}{idx_serie}][name]`.

        Parameters
        ----------
        name : str
            Attribute name.
        value : Any
            Attribute value to set.
        """
        where = f"{self._results_name}_{self._serie_name}{self.idx_serie}"
        self.data_processed[where].attrs[name] = value

    def get_nexus_filename(self) -> str:
        """Generate output NeXus full file name."""
        return os.path.join(self.cfg.data_directory, self.cfg.expid + ".nx5")

    def consolidate(self) -> Self:
        """
        Add supplementary NeXus entries from configuration file.

        The "nx" section of the configuration is read to add new groups and datasets to
        provide further details about the sample and experiment.
        """
        for nxk, nxv in self.cfg.nx.items():
            # [nx.xyz] section
            nxclass = "NX" + nxk
            nxgroup = nx.NXgroup(nxclass=nxclass)
            for k, v in nxv.items():
                # write dataset
                nxgroup[k] = v
            if nxk in self.nxentry:
                del self.nxentry[nxk]
            self.nxentry[nxk] = nxgroup

        return self

    def resolve_nexus_links(self, path: str = "/") -> Self:
        """
        Replace placeholders in the NXroot group by actual links to other datasets.

        To enable automatic plotting by NeXus programs (such as `nexpy`), the datasets
        plotted one versus the other need to be in the same groups. Common x-axis, such
        as time or magnetic field, are linked to the actual datasets found in
        `data_processed[results]`.

        In practice, this method recursively scans all datasets to find datasets that
        are strings and contains the pre-defined placeholders formatted like so :

        "!link to:path/to/dataset"

        and are replaced with a NXlink targeting the specified dataset.
        """
        nxgrp = self.nxroot[path]
        for obj in nxgrp.values():
            if isinstance(obj, nx.NXfield):
                # is a dataset, not a group
                if isinstance(obj.nxdata, bytes):
                    # dataset is a string
                    val = obj.nxdata.decode("utf-8")
                elif isinstance(obj.nxdata, str):
                    # dataset is a string
                    val = obj.nxdata
                else:
                    # not a string, continue scanning
                    continue
                if val.startswith("!link to:"):
                    # begins with the placeholder flag, read the target
                    target = val.split("!link to:")[1]
                else:
                    # not beginning with the link placeholder
                    continue
                # replace placeholder with an actual link
                self.nxroot[obj.nxpath] = nx.NXlink(
                    self.nxroot[self.nxentry.nxname][target]
                )

            else:
                # is a group, continue scanning within the group
                self.resolve_nexus_links(path=obj.nxpath)

        return self

    def release_nexus_file(self) -> None:
        """
        Uncouple NeXus data structure and the NeXus HDF5 file.

        It *should* prevent changes made to the object be reflected in the file.
        """
        self.nxroot.nxfile.clear_lock()
        self.nxroot.nxfile.close()
        self.nxroot._filename = None
        self.nxroot._file = None

    def plot_field(self) -> plt.Figure | None:
        """
        Plot magnetic field.

        Pull data from `data_processed[results]["magfield"]` and plot it against the
        time vector.
        """
        if not self.get_data_processed(
            "magfield_time", checkonly=True
        ) or not self.get_data_processed("magfield", checkonly=True):
            logger.warning("The magnetic field was not computed, cannot plot field")
            return

        fig = plt.figure()
        plt.plot(
            self.get_data_processed("magfield_time"),
            self.get_data_processed("magfield"),
        )
        plt.xlabel("time (s)")
        plt.ylabel("B (T)")
        plt.show()

        return fig

    def _init(self) -> None:
        """
        Initialize the object after loading a configuration file.

        This method should be executed either *after* loading the configuration file or
        *before* loading a NeXus file.

        The Data objects from the `data` modules and the NeXus structure are initialized
        here.
        """
        # Get parameters and metadata
        self.measurements = [*self.cfg.measurements.keys()]
        attr_raw = self.cfg.nexus.groups["data"]
        attr_processed = self.cfg.nexus.groups["analysis"]
        name_entry = self.cfg.nexus.groups["root"]["name"]

        # Prepare attributes storing the (meta)data.
        self.metadata = dict()
        self.data_raw = DataRaw(attr=attr_raw)
        self.data_processed = DataProcessed(
            program=self._program_name,
            attr=attr_processed,
            results_name=self._results_name,
            serie_name=self._serie_name,
        )
        # NeXus Entry
        self.nxentry = nx.NXentry(name=name_entry)
        self.nxentry.attrs["default"] = "processed"
        self._fill_nexus_entry()
        # NeXus Root
        self.nxroot = nx.NXroot(self.nxentry)
        self.nxroot.attrs["default"] = name_entry

    def _load_pickup(
        self,
        filename: Path | str,
        precision: int,
        endian: Literal["<", ">"] = "<",
        order: Literal["F", "C"] = "F",
        nseries: int = 1,
        index: int = 0,
    ) -> np.ndarray:
        """
        Load the pickup data from a binary file.

        The array is reshaped given the number series (by default, 1).

        Parameters
        ----------
        filename : Path | str
            Full path to the pickup binary file.
        precision : int
            Byte precision.
        endian : {"<", ">"}, optional
            "<" for little endian, ">" for big endian. Default is "<".
        order : {"F", "C"}, optional
            Array order, "F" for Fortran, "C" for C. Default is "F".
        nseries : int, optional
            Number of pickups time series in the binary file. Default is 1.
        index : int, optional
            Index of the time serie to read, 0-based. Default is 0 (first).

        Returns
        -------
        pickup : np.ndarray
            Pickup voltage time serie.
        """
        # Read data
        data = self.load_bin(filename, precision, endian)

        # Reshape and get required time serie
        data = data.reshape((-1, nseries), order=order)[:, index].astype(float).copy()

        return data

    @staticmethod
    def _make_fake_pickup(npoints: int, start_at: float = 0.0) -> np.ndarray:
        """
        Generate a fake pickup signal.

        The generated signal is a simple linear function.

        Parameters
        ----------
        npoints : int
            Number of data points.
        start_at : float, optional
            Starting point, default is 0.
        """
        return np.linspace(start_at, npoints, npoints, endpoint=False)

    def _compute_field(
        self,
        surface: float = 1.0,
        method: str = "trapz",
    ) -> Self:
        """
        Compute magnetic field from pickup coil data.

        Wraps the `sp.integrate_pickup()` function. Resulting field is stored as
        `magfield` in `data_processed`, with the corresponding time vector as
        `magfield_time`.

        Requires `pickup` and `pickup_time` in `data_raw`. If pickup signal is empty, a
        synthetic one is generated.

        Parameters
        ----------
        surface : float
            Pickup coil surface in m².
        method : str, optional
            Integration method. Default is "trapz" (which is the only one supported).
        """
        if ("pickup_time" not in self.data_raw) or ("pickup" not in self.data_raw):
            logger.error("Pickup signal was not loaded, can't compute magnetic field")
            return self

        pu_time = self.get_data_raw("pickup_time")
        pu_signal = self.get_data_raw("pickup")

        # Check if we need to simulate a signal
        if pu_time.shape == (0,) or pu_signal.shape == (0,):
            # We need to simulate a pickup signal
            if self.get_data_processed("time_exp", checkonly=True):
                logger.info("There is no pickup signal, simulating one")
                self.create_fake_pickup()
                # Reload
                pu_time = self.get_data_raw("pickup_time")
                pu_signal = self.get_data_raw("pickup")
            else:
                # Required data missing
                logger.warning(
                    "There is no pickup signal, main data need to be loaded first to "
                    "simulate a field"
                )
                return self

        # Integrate and store
        logger.info("Computing magnetic field...")
        self.set_data_processed(
            "magfield", sp.integrate_pickup(pu_time, pu_signal, surface, method=method)
        )
        self.set_data_processed("magfield_time", self.get_data_raw("pickup_time"))

        # Verbose
        maxfield = self.get_data_processed("magfield").max()
        logger.info(f"Done, max. field = {maxfield:2.2f}T")

        return self

    def _fill_nexus_entry(self) -> Self:
        """
        Fill the NXentry with the Data objects.

        'data_processed' and 'data_raw' are added as NXsubentry.
        """
        # Add data
        self.nxentry["processed"] = nx.NXsubentry(self.data_processed)
        self.nxentry["processed"].attrs["default"] = "analysis"
        self.nxentry["raw"] = nx.NXsubentry(self.data_raw)
        self.nxentry["raw"].attrs["default"] = "data"

        return self

    def _set_data(self, where: str, name: str, value: Any) -> None:
        """
        Set data in the DataProcessed object.

        Dataset will be written as an NXfield in `data_processed[where][name]`. If the
        dataset already exists, it is replaced (modifying its type and shape
        accordingly).

        If the group name or the dataset name are present in the 'nexus' section of the
        configuration, the attributes are written as well.

        Parameters
        ----------
        where : str
            Location of the dataset in the DataProcessed object.
        name : str
            Dataset name.
        value : Any
            Dataset.
        """
        # Put data
        if name in self.data_processed[where]:
            self.data_processed[where][name].replace(value)
        else:
            self.data_processed[where][name] = value

        # Set group attributes
        if where in self.cfg.nexus.groups:
            for k, v in self.cfg.nexus.groups[where].items():
                self.data_processed[where].attrs[k] = v
        # Set dataset attributes
        if name in self.cfg.nexus.datasets:
            for k, v in self.cfg.nexus.datasets[name].items():
                self.data_processed[where][name].attrs[k] = v

    def _get_data(self, where: str, name: str, checkonly: bool = False) -> Any:
        """
        Get data from the DataProcessed object.

        The data is pulled from `data_processed[where][name]`. If `checkonly` is True,
        only check if the dataset exists.

        Note : the returned value is a *view*, if it is modified, the underlying data
        is modified as well.

        Parameters
        ----------
        where : str
            Location of the dataset in the DataProcessed object.
        name : str
            Dataset name.
        checkonly : bool, optional
            Perform a key existence check without returning the value. Default is False.
        """
        if not checkonly:
            return self.data_processed[where][name].nxdata
        else:
            if where not in self.data_processed:
                return False
            if name not in self.data_processed[where]:
                return False
            else:
                return True

    def _remove_data(self, where: str, name: str) -> None:
        """
        Remove data from the DataProcessed object.

        Parameters
        ----------
        where : str
            Location of the dataset in the DataProcessed object.
        name : str
            Dataset name.
        """
        if self._get_data(where, name, checkonly=True):
            del self.data_processed[where][name]

    def _check_field_computed(self) -> bool:
        """Check if magnetic field was computed."""
        return self.get_data_processed(
            "magfield_time", checkonly=True
        ) and self.get_data_processed("magfield", checkonly=True)

    def _check_field_aligned(self) -> bool:
        """Check if the magnetic field and data time bases are aligned."""
        if not self.get_data_processed("time_exp", checkonly=True):
            return False

        if not self._check_field_computed():
            return False
        return (
            self.get_data_processed("time_exp").size
            == self.get_data_processed("magfield_time").size
        )



"""The Configuration widget for pyuson."""

from functools import partial

from PyQt6.QtCore import pyqtSignal, pyqtSlot
from pyqtgraph.parametertree import Parameter

from pymagnetos.core.gui.widgets import BaseConfigurationWidget

from ._param_content import ParamContent


class ConfigurationWidget(BaseConfigurationWidget):
    """
    Initialize a PyQtGraph ParameterTree.

    Signals
    -------
    sig_find_f0 : emits when the "Find f0" button is clicked.
    sig_demodulate : emits when the "Demodulate" button is clicked.
    """

    sig_find_f0 = pyqtSignal()
    sig_demodulate = pyqtSignal()

    def __init__(self, param_content: type[ParamContent]) -> None:
        super().__init__(param_content)
        self.add_extra_parameter()

    def init_demodulation_tree(self) -> None:
        """Create ParameterTree for demodulation parameters."""
        if not hasattr(self._param_content, "children_demodulation"):
            raise ValueError(
                "Could not add the demodulation parameter tree, not in digital mode"
            )
        self.demodulation_parameters = Parameter.create(
            name="Demodulation",
            type="group",
            children=self._param_content.children_demodulation,
        )
        self.host_parameters.addChild(self.demodulation_parameters)

        # Store buttons as the others
        self.button_findf0 = self.demodulation_parameters.child("find_f0")
        self.button_findf0.setOpts(enabled=False)
        self.button_demodulate = self.demodulation_parameters.child("demodulate")
        self.button_demodulate.setOpts(enabled=False)

        self.connect_to_signals_demodulation()

    def add_extra_parameter(self) -> None:
        """Add extra parameters to the tree, only for the GUI."""
        # Show the RF frequency read in the metadata file
        rf_freq = Parameter.create(
            name="rf_frequency",
            type="float",
            readonly=True,
            value=-1,
            suffix="Hz",
            siPrefix=True,
            title="RF frequency in metadata",
        )
        self.host_parameters.addChild(rf_freq)

    def connect_to_signals_demodulation(self) -> None:
        """Additionnal connections for demodulation."""
        self.button_findf0.sigActivated.connect(self.sig_find_f0.emit)
        self.button_demodulate.sigActivated.connect(self.sig_demodulate.emit)

        for p in self.demodulation_parameters:
            self.demodulation_parameters.child(p.name()).sigValueChanged.connect(
                partial(self.send_signal_on_change, "demodulation", p.name())
            )

    @pyqtSlot(str, str)
    def send_signal_on_change(
        self,
        context: str,
        param_name: str,
    ) -> None:
        """
        Send a signal when a parameter in the tree is changed.

        The context (section of the tree - "parameters", "settings", ...), parameter
        name, and value is emitted.
        """
        if context == "demodulation":
            current_value = self.demodulation_parameters[param_name]

            # Parse it if needed
            if param_name in self.parameters_to_parse:
                # special case that needs to be parsed
                current_value = self.get_numbers_from_text(current_value)

            # Send the signal
            self.sig_parameter_changed.emit(context, param_name, current_value)
        else:
            # Use parent handler
            super().send_signal_on_change(context, param_name)

    def enable_buttons(self) -> None:
        super().enable_buttons()
        if hasattr(self, "button_findf0"):
            self.button_findf0.setOpts(enabled=True)
        if hasattr(self, "button_demodulate"):
            self.button_demodulate.setOpts(enabled=True)

    def disable_buttons(self) -> None:
        super().disable_buttons()
        if hasattr(self, "button_findf0"):
            self.button_findf0.setOpts(enabled=False)
        if hasattr(self, "button_demodulate"):
            self.button_demodulate.setOpts(enabled=False)



"""
The Configuration panel widget.

Contains the pyqtgraph ParameterTree that represents a configuration file, hosting all
the parameters and settings for the analysis.

It requires a ParamContent class which lists all the parameters to include in the tree.
See the `_param_content.py` module.
"""

import re
from collections.abc import Iterable
from functools import partial

from PyQt6 import QtWidgets
from PyQt6.QtCore import pyqtSignal, pyqtSlot
from pyqtgraph.parametertree import Parameter, ParameterTree

from ._param_content import BaseParamContent


class BaseConfigurationWidget(QtWidgets.QWidget):
    """
    Initialize a PyQtGraph ParameterTree.

    pyqtSignals
    -------
    sig_file_changed : emits when the file parameter changed.
    sig_expid_changed : emits when the "Reload data" button is clicked or the experiment
        ID parameter changed.
    sig_autoload_changed : emits when the autoload checkbox in the File section is
        changed.
    sig_reload_config : emits when the "Reload config" button is clicked.
    sig_save_config : emits when the "Save config" button is clicked.
    sig_parameter_changed : emits when any other parameter in the tree is changed. Emits
        the context (tree section, "settings" or "parameters"), the parameter name and
        its value.
    """

    sig_file_changed = pyqtSignal()
    sig_expid_changed = pyqtSignal()
    sig_autoload_changed = pyqtSignal()

    sig_save_config = pyqtSignal()
    sig_reload_config = pyqtSignal()

    sig_parameter_changed = pyqtSignal(str, str, object)

    def __init__(self, param_content: type[BaseParamContent]) -> None:
        super().__init__()
        self._param_content = param_content
        self.parameters_to_parse = self._param_content.PARAMS_TO_PARSE

        # Create trees and buttons
        self.init_files_tree()
        self.init_buttons()
        self.init_configuration_tree()

        # Connect
        self.connect_to_signals()

        # Create layout
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.files_tree, stretch=3)
        layout.addWidget(self.widget_buttons, stretch=1)
        layout.addWidget(self.config_tree, stretch=10)

        self.setLayout(layout)

    def init_files_tree(self) -> None:
        """Create the File section."""
        # Files section
        self.files_parameters = Parameter.create(
            name="Files", type="group", children=self._param_content.children_files
        )
        self.files_tree = ParameterTree(showHeader=False)
        self.files_tree.setParameters(self.files_parameters)

    def init_buttons(self) -> None:
        """Create and connect the file-related buttons."""
        # Create buttons
        self.button_reload_data = QtWidgets.QPushButton("Reload data", self)
        self.button_reload_config = QtWidgets.QPushButton("Reload config", self)
        self.button_save_config = QtWidgets.QPushButton("Save config", self)

        # Connect
        self.button_reload_data.clicked.connect(self.sig_expid_changed.emit)
        self.button_reload_config.clicked.connect(self.sig_reload_config.emit)
        self.button_save_config.clicked.connect(self.sig_save_config)

        # Create a layout
        layout_buttons = QtWidgets.QHBoxLayout(self)
        layout_buttons.addWidget(self.button_reload_data)
        layout_buttons.addWidget(self.button_reload_config)
        layout_buttons.addWidget(self.button_save_config)
        self.widget_buttons = QtWidgets.QWidget(self)
        self.widget_buttons.setLayout(layout_buttons)

    def init_configuration_tree(self) -> None:
        """Create the ParameterTree, with the Parameters and Settings sections."""
        # Parameters section
        self.param_parameters = Parameter.create(
            name="Parameters",
            type="group",
            children=self._param_content.children_parameters,
        )
        # Settings section
        self.settings_parameters = Parameter.create(
            name="Settings",
            type="group",
            children=self._param_content.children_settings,
        )

        # Host Tree
        self.host_parameters = Parameter.create(
            name="Configuration",
            type="group",
            children=[
                self.param_parameters,
                self.settings_parameters,
            ],
        )
        self.config_tree = ParameterTree(showHeader=False)
        self.config_tree.setParameters(self.host_parameters)

    def get_numbers_from_text(self, text: str) -> list[float]:
        """Parse input text as list of numbers."""
        if isinstance(text, str):
            pattern = r"-?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?"
            return [float(el) for el in re.findall(pattern, text)]
        else:
            raise TypeError(f"Can't parse numbers from type {type(text)}")

    def get_text_from_numbers(
        self, inds: Iterable[float | int], precision: int = 10
    ) -> str:
        """
        Convert input list of numbers as text.

        There is a given precision, the output string should be used for display purpose
        only.
        """
        if isinstance(inds, Iterable):
            return ", ".join(
                f"{int(e)}" if int(e) == e else f"{e:<.{precision}g}" for e in inds
            )
        else:
            raise TypeError(f"Can't parse numbers nor text from type {type(inds)}")

    def connect_to_signals(self) -> None:
        """Connect changes to signals."""
        self.files_parameters.child("file").sigValueChanged.connect(
            self.sig_file_changed.emit
        )
        self.files_parameters.child("expid").sigValueChanged.connect(
            self.sig_expid_changed.emit
        )
        self.files_parameters.child("autoload").sigValueChanged.connect(
            self.sig_autoload_changed.emit
        )

        for p in self.param_parameters:
            self.param_parameters.child(p.name()).sigValueChanged.connect(
                partial(self.send_signal_on_change, "parameters", p.name())
            )

        for p in self.settings_parameters:
            self.settings_parameters.child(p.name()).sigValueChanged.connect(
                partial(self.send_signal_on_change, "settings", p.name())
            )

    @pyqtSlot(str, str)
    def send_signal_on_change(
        self,
        context: str,
        param_name: str,
    ) -> None:
        """
        Send a signal when a parameter in the tree is changed.

        The context (section of the tree - "parameters", "settings", ...), parameter
        name, and value is emitted.
        """
        # Collect the parameter value
        match context:
            case "parameters":
                current_value = self.param_parameters[param_name]
            case "settings":
                current_value = self.settings_parameters[param_name]
            case _:
                raise ValueError(f"Unknown configuration section : {context}")

        # Parse it if needed
        if param_name in self.parameters_to_parse:
            # special case that needs to be parsed
            current_value = self.get_numbers_from_text(current_value)

        # Send the signal
        self.sig_parameter_changed.emit(context, param_name, current_value)

    def enable_buttons(self) -> None:
        """Enable all buttons."""
        self.button_reload_data.setEnabled(True)
        self.button_reload_config.setEnabled(True)
        self.button_save_config.setEnabled(True)

    def disable_buttons(self) -> None:
        """Disable all buttons."""
        self.button_reload_data.setEnabled(False)
        self.button_reload_config.setEnabled(False)
        self.button_save_config.setEnabled(False)



"""Worker that connects to an TDOProcessor object."""

from pathlib import Path

from PyQt6.QtCore import pyqtSignal as Signal
from PyQt6.QtCore import pyqtSlot as Slot

from pymagnetos.core.gui import BaseWorker
from pymagnetos.pytdo import TDOProcessor


class Worker(BaseWorker):
    """
    Wrapper around the TDOProcessor class.

    It exposes the EchoProcessor methods as slots and emits a signal when the task is
    finished.

    Additional signals
    ------------------
    sig_crop_finished : emits when the signal is time-cropped.
    sig_offset_finished : emits when the pickup/signal offset is applied.
    sig_tdocsv_finished : emits when the TDO signal is saved as CSV.
    sig_rescsv_finished : emits when the analysis is saved as CSV.
    sig_load_csv_finished : emits when a CSV file is loaded.
    """

    _type_processor: type[TDOProcessor] = TDOProcessor
    proc: TDOProcessor

    sig_crop_finished = Signal()
    sig_offset_finished = Signal()
    sig_tdocsv_finished = Signal()
    sig_rescsv_finished = Signal()
    sig_load_csv_finished = Signal()

    @Slot()
    def crop_signal(self):
        """Crop signal."""
        self.proc.crop_signal(force=True)
        self.sig_crop_finished.emit()

    @Slot()
    def time_offset(self):
        """Apply time offset on experiment time vector."""
        self.proc.apply_time_offset()
        self.sig_offset_finished.emit()

    @Slot(str)
    def save_tdo_csv(self, fname: str | Path):
        """Save TDO signal as CSV."""
        self.proc.export_tdo_csv(filename=fname)
        self.sig_tdocsv_finished.emit()

    @Slot(str)
    def save_results_csv(self, fname: str | Path):
        """Save results as CSV."""
        self.proc.export_results_csv(filename_prefix=fname)
        self.sig_rescsv_finished.emit()

    @Slot(str)
    def load_csv_file(self, fname: str):
        """Save data as CSV file."""
        self.proc.load_csv(fname)
        self.sig_load_csv_finished.emit()



"""A generic Config base class to load TOML configuration file, using Pydantic."""

import logging
import tomllib
from pathlib import Path
from typing import Any, Literal

import tomlkit
from pydantic import BaseModel

from .config_models import File
from .utils import merge_dict_nested, remove_duplicate_values, strip_none_dict

# Only attributes than can be set for the Config class, otherwise they're set in the
# underlying Pydantic model
SELF_ATTRIBUTES = ("_cfg", "_model", "_user_file", "_default_file", "_overrides")

logger = logging.getLogger(__name__)


class BaseConfig:
    """
    Load and store parameters from user-defined configuration TOML file.

    It uses an underlying Pydantic model stored in the `_cfg` attribute. Setting an
    attribute will actually set it in the underlying model (raising an error if it is
    not defined).
    """

    def __init__(
        self,
        model: type[BaseModel],
        user_file: str | Path | None = None,
        default_file: str | Path | None = None,
        no_config: bool = False,
        **overrides: Any,
    ) -> None:
        """
        Load and store parameters from user-defined configuration TOML file.

        It requires a fully-defined Pydantic model.

        If `user_file` is a JSON file, the file is read and given to the Pydantic
        `model_validate_json()` method and the default file and overrides are ignored as
        it is assumed that a JSON configuration file comes from a previously serialized
        Config object.

        Otherwise, the configuration is set from, in this order of preference :
        1. **overrides
        2. Configuration file (`user_file`)
        3. Default configuration file (`default_file`). If not provided, the one bundled
        with the package is used.

        To load a JSON string (not a file), use the `loads()` class method.

        Parameters
        ----------
        model : pydantic.BaseModel
            A fully-defined (custom) Pydantic BaseModel.
        user_file : str or Path or None
            Path to a user configuration file.
        default_file : str, Path or None
            Path to the default configuration file used as fallback for missing
            parameters in the user file.
        no_config : bool, optional
            Initialize an empty Config object, with just the Pydantic model. Default is
            False.
        **overrides : overriding keyword arguments
            Keyword arguments that overwrite parameters from other sources.
        """
        self._model = model

        if no_config:
            self._default_file = None
            self._user_file = None
            self._overrides = overrides
        else:
            self._default_file = Path(default_file) if default_file else None
            self._user_file = Path(user_file) if user_file else None
            self._overrides = overrides
            self.load()
            self._patch_old_version()
            self._resolve_data_directory()
            self.build_filenames()
            self._greeting()

    def __getattr__(self, name: str, /) -> Any:
        """
        Return `name` from the underlying Pydantic model.

        This magic method is called only when an attribute is not found.
        """
        return getattr(self._cfg, name)

    def __setattr__(self, name: str, value: Any, /) -> None:
        """
        Set `name` to `value` in the underlying Pydantic model.

        Only a restricted list of attributes can be set in the Config object itself (see
        `SELF_ATTRIBUTES`).
        """
        if name in SELF_ATTRIBUTES:
            super().__setattr__(name, value)
        else:
            setattr(self._cfg, name, value)

    def load(self, merge_measurement_dict: bool = False) -> None:
        """
        Load a configuration file.

        Default and user configuration files are merged, with priority to the user
        configuration. The "measurement" section is a special case that can optionally
        be entirely replaced by the user-provided dictionary.

        Parameters
        ----------
        merge_measurement_dict : bool, optional
            If False (default), the "measurement" section of the configuration file is
            replaced by the content of the user-provided file instead of merging with
            the default file.
        """
        if self._default_file:
            default_config = tomllib.loads(self._default_file.read_text())
        else:
            default_config = dict()

        if isinstance(self._user_file, Path):
            if self._user_file.suffix == ".json":
                # Use Pydantic built-in method for JSON
                self._cfg = self._model.model_validate_json(self._user_file.read_text())
                return

            # Do not use tomlkit as it does not play nicely with Pydantic
            user_config = tomllib.loads(self._user_file.read_text())
            new_values = merge_dict_nested(user_config, default_config)

            # Handle the "measurements" section : user-provided and default entries were
            # merged
            if len(user_config.get("measurements", [])) and not merge_measurement_dict:
                # Use user-provided entries only
                new_values["measurements"] = user_config["measurements"]
            else:
                # Merge with default, but remove duplicate values, keeping the
                # user-provided entries
                new_values["measurements"] = remove_duplicate_values(
                    new_values["measurements"], default_config["measurements"]
                )

        else:
            new_values = default_config

        if len(self._overrides) > 0:
            newer_values = merge_dict_nested(self._overrides, new_values)
        else:
            newer_values = new_values

        self._cfg = self._model(**newer_values)

    def _model_validate_json(self, json_data: str | bytes, **kwargs):
        """Read and set the configuration from a JSON string."""
        self._cfg = self._model.model_validate_json(json_data, **kwargs)
        self._greeting()

    def _greeting(self):
        logger.info(
            f"Configuration loaded for experiment: '{self.expid}'."
            f" Data directory: {self.data_directory}"
        )

    @classmethod
    def loads(cls, json_data: str | bytes, **kwargs):
        """
        Load a JSON-formatted string as the Config.

        Subclasses must implement this with the `_loads()` method and the relevant
        Pydantic model.
        """
        raise NotImplementedError(
            "Subclasses must implement this method with Pydantic model."
        )

    @classmethod
    def _loads(cls, model: type[BaseModel], json_data: str | bytes, **kwargs):
        """Create a Config object from a JSON string."""
        cfg = cls(model, no_config=True)
        cfg._model_validate_json(json_data, **kwargs)
        return cfg

    def build_filenames(self):
        """
        Build data file paths with the parameters found in `files`.

        Files names are built like this :
        {data_directory}/{expid}{ext}
        For oscilloscope data saved as Tektronix WFM files :
        + {data_directory}/{expid}_ch{n}{ext}
        """
        self.filenames = self._build_filenames(self.files)

    def resolve_nexus(self, serie_name: str):
        """
        Patch the `[nexus.groups]` section of the configuration.

        Replace "serie" in keys with `serie_name`, and if no name is specified for the
        main NXentry, set it to the dataset name.

        Parameters
        ----------
        serie_name : str
            Name of the series.
        """
        # Set the NXroot object name if not specified
        if not self.nexus.groups["root"]["name"]:
            self.nexus.groups["root"]["name"] = self.expid

        # Replace "serie" in key names with `serie_name`
        groups = self.nexus.groups.copy()
        for key in groups:
            self.nexus.groups[key.replace("serie", serie_name)] = self.nexus.groups.pop(
                key
            )

    def write(
        self,
        output_file: str | Path,
        format: Literal["guess", "toml", "json"] = "guess",
        overwrite: bool = False,
    ) -> bool:
        """
        Save current configuration to file.

        If the target file exists and `overwrite` is False (default), the file is not
        written. By default, the write mode is guessed from the file extension (json or
        toml), to force a mode, use the `format` keyword argument.

        Parameters
        ----------
        output_file : str or Path
            Path to the output file.
        format : {"guess", "json", "toml"}, optional
            Output file format. `"guess"` infers from file extension, this is the
            default.
        overwrite: bool, optional
            Whether to overwrite the output file if it exists. Default is False.
        """
        output_file = Path(output_file)
        # Check there is something to do
        if output_file.is_file() and not overwrite:
            logger.warning(f"{output_file.name} already exists, not saving.")
            return False
        if format not in ("guess", "toml", "json"):
            logger.error("File format not allowed, choose 'guess', 'toml' or 'json'.")
            return False

        # Determine file format
        if format == "guess":
            if output_file.suffix.endswith(".toml"):
                format = "toml"
            elif output_file.suffix.endswith(".json"):
                format = "json"
            else:
                logger.error(
                    f"Couldn't infer file format from file name: {output_file.name}"
                )
                return False

        try:
            if format == "toml":
                self._save_toml(output_file)
            elif format == "json":
                self._save_json(output_file)
            logger.info(f"Configuration saved at {output_file}.")
            return True
        except Exception as e:
            logger.error(f"Failed to save configuration file ({e}).")
            return False

    def model_dump_json(self, *args, **kwargs) -> str:
        """
        Dump the model in a JSON string.

        Just a wrapper for the underlying Pydantic model `model_dump_json()` method.
        Because of the `__getattr()__` magic method, this method is not mandatory. It is
        still implemented for clarity.
        """
        return self._cfg.model_dump_json(*args, **kwargs)

    def _patch_old_version(self):
        """
        Further adjustments to adapt to previous configuration file version.

        The 'filenames' section should be used to specify direct path to the data files.
        + 'base' is replaced by a more explicit 'expid' and set at the root of the file,
            outside of any section.
        + 'data_directory' should be set at the root of the file outside of any section.
        """
        if "base" in self.filenames:
            logger.warning(
                "Setting 'base' in the 'filenames' section is deprecated. "
                "Set the 'expid' parameter at the top of the file instead."
            )
            self.expid = str(self.filenames.pop("base"))
        if "data_directory" in self.filenames:
            logger.warning(
                "Setting 'data_directory' in the 'filenames' section is deprecated. "
                "Set the 'data_directory' parameter at the top of the file instead."
            )
            self.data_directory = Path(self.filenames.pop("data_directory"))

    def _resolve_data_directory(self):
        """
        Resolve the data directory entry from the configuration file.

        If it is "." or omitted, the directory where the file is is used. Otherwise,
        the location specified by the user is used.
        """
        if self.data_directory in (None, Path(".")) and self._user_file is not None:
            self.data_directory = self._user_file.parent
        elif isinstance(self.data_directory, str):
            self.data_directory = Path(self.data_directory)

    def _build_filenames(self, files_dic: dict[str, File]) -> dict[str, Path]:
        """
        Build file names based on configuration.

        Files names are built like this :
        {data_directory}/{expid}{ext}
        For oscilloscope data saved as Tektronix WFM files :
        + {data_directory}/{expid}_ch{n}{ext}

        Parameters
        ----------
        files_dic : dict
        """
        # Resolve the various files that should be entries in the [files] section
        datadir = Path(self.data_directory)  # for convenience in this function
        filenames_dic = dict()
        for entry in files_dic:
            # Get file extension
            ext = files_dic[entry].ext
            if ext.endswith("wfm"):
                # Special case for oscilloscope : one file per channel, add a keyword
                filenames_dic[entry] = datadir / f"{self.expid}_!CHANNELID{ext}"

            else:
                filenames_dic[entry] = datadir / f"{self.expid}{ext}"

        return filenames_dic

    def _save_json(self, output_file: Path):
        """Save current configuration to a JSON file."""
        output_file.write_text(self.model_dump_json(), encoding="utf8")

    def _save_toml(self, output_file: Path):
        """Save current configuration to a TOML file."""
        output_file.write_text(
            tomlkit.dumps(strip_none_dict(self._cfg.model_dump(mode="json"))),
            encoding="utf8",
        )



"""Worker that connects to an EchoProcessor object."""

from PyQt6.QtCore import pyqtSignal as Signal
from PyQt6.QtCore import pyqtSlot as Slot

from pymagnetos.core.gui import BaseWorker
from pymagnetos.pyuson import EchoProcessor


class Worker(BaseWorker):
    """
    Wrapper around the EchoProcessor object.

    It exposes the EchoProcessor methods as slots and emits a signal when the task is
    finished.

    Additional signals
    ------------------
    sig_rolling_finished : emits when rolling average is finished.
    sig_find_f0_finished : emits f0 (float) when the center frequency is found.
    sig_export_csv_finished : emits when the results are exported to CSV.
    """

    proc: EchoProcessor
    _type_processor: type[EchoProcessor] = EchoProcessor

    sig_rolling_finished: Signal = Signal()
    sig_find_f0_finished: Signal = Signal(float)
    sig_export_csv_finished: Signal = Signal()

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # Setup methods arguments
        self._after_load_align = True
        self._kwargs_load = {"scale": True}
        self._kwargs_preprocess = {"progress_emitter": self.sig_preprocess_progress}
        self._kwargs_analysis = {}

    @Slot()
    def rolling_average(self) -> None:
        """Apply moving average."""
        self.proc.rolling_average()
        self.sig_rolling_finished.emit()

    @Slot()
    def find_f0(self) -> None:
        """Find F0."""
        self.proc.find_f0()
        self.sig_find_f0_finished.emit(self.proc.cfg.demodulation.f0)

    @Slot(tuple)
    def batch_process(
        self,
        args: tuple,
    ) -> None:
        """Batch-process several datasets."""
        self.proc.batch_process(
            *args,
            demodulation_progress_emitter=self.sig_preprocess_progress,
            batch_progress_emitter=self.sig_batch_progress,
        )

        # Emit signals
        self.sig_load_data_finished.emit()
        if self.proc.is_digital:
            self.sig_preprocess_finished.emit()
        self.sig_analysis_finished.emit()
        self.sig_batch_finished.emit()

    @Slot(str, bool)
    def export_as_csv(self, fname: str, to_cm: bool) -> None:
        """Export results as CSV."""
        self.proc.export_csv(fname=fname, to_cm=to_cm)
        self.sig_export_csv_finished.emit()



"""
The linx subpackage, a library for digital lock-in.

Provides
--------
- `LinxConfig` : The `Config` object specific to digital lock-in.
- `LinxProcessor` : the `Processor` object specific digital lock-in.
- `config_models` : the `pydantic` models to define the configuration requirements.
- `gui` : a graphical application. It needs to be imported explicitely.
"""

from . import config_models
from ._config import LinxConfig
from ._linxprocessor import LinxProcessor

__all__ = ["LinxConfig", "LinxProcessor", "config_models"]



"""Application for TDO analysis."""

import importlib.metadata
from collections.abc import Sequence
from pathlib import Path

from PyQt6 import QtGui, QtWidgets
from PyQt6.QtCore import pyqtSignal, pyqtSlot

from pymagnetos.core.gui import BaseMainWindow
from pymagnetos.core.gui.widgets import BaseBatchProcessingWidget, FileBrowserWidget

from .widgets import ButtonsWidget, ConfigurationWidget, GraphsWidget, ParamContent
from .worker import Worker

ICON_PATH = str(Path(__file__).parent / "assets" / "icon.png")
REGEXP_EXPID_SEPARATORS = r"[_-]"
PROGRAM_NAME = "pymagnetos"
OUT_FORMAT = (".txt", ".csv", ".tsv", ".out")
LOG_LEVEL = "INFO"
VERSION = importlib.metadata.version(PROGRAM_NAME)

# List of parameter names that should trigger an analysis
PARAM_ANALYSIS_TRIGGER = [
    "poly_deg",
    "npoints_interp_inverse",
    "fft_pad_mult",
    "max_bfreq",
]


class MainWindow(BaseMainWindow):
    """
    A graphical user interface application for TDO analysis.

    The `MainWindow` class defines the main thread with the front-end user interface.
    """

    sig_worker_crop = pyqtSignal()
    sig_worker_offset = pyqtSignal()
    sig_worker_tdocsv = pyqtSignal(str)
    sig_worker_rescsv = pyqtSignal(str)
    sig_worker_loadcsv = pyqtSignal(str)

    worker: Worker

    def __init__(self):
        # Register the widgets
        self._type_wbatch = BaseBatchProcessingWidget
        self._type_wbuttons = ButtonsWidget
        self._type_wconfiguration = ConfigurationWidget
        self._type_wfiles = FileBrowserWidget
        self._type_wgraphs = GraphsWidget
        self._param_content = ParamContent
        self._type_worker = Worker

        super().__init__()

        # Initialize window
        self.setGeometry(300, 300, 900, 450)
        self.setWindowTitle("TDO Analyzer")

        self.logger.info(f"Running pymagnetos.pytdo v{VERSION}")

    def init_buttons(self):
        """Create the main buttons."""
        super().init_buttons()

        # Connect
        self.wbuttons.sig_extract.connect(self.preprocess)
        self.wbuttons.sig_analyse.connect(self.run_analysis)
        self.wbuttons.sig_tdocsv.connect(self.save_tdo_csv)
        self.wbuttons.sig_rescsv.connect(self.save_results_csv)

    def init_plots(self):
        """Create the graphs area."""
        super().init_plots()

        # Connect ROIs
        self.wgraphs.sig_roi1_changed.connect(self.handle_roi1_changed)
        self.wgraphs.sig_roi2_changed.connect(self.handle_roi2_changed)

    def init_log(self):
        """Setup logging."""
        super()._init_log(PROGRAM_NAME, log_level=LOG_LEVEL)

    def connect_worker(self):
        """Connect signals to the worker slots."""
        super().connect_worker()

        # Cropping
        self.sig_worker_crop.connect(self.worker.crop_signal)
        self.worker.sig_crop_finished.connect(self.crop_signal_finished)

        # Time offset
        self.sig_worker_offset.connect(self.worker.time_offset)
        self.worker.sig_offset_finished.connect(self.time_offset_finished)

        # Save TDO signal as CSV
        self.sig_worker_tdocsv.connect(self.worker.save_tdo_csv)
        self.worker.sig_tdocsv_finished.connect(self.save_tdo_csv_finished)

        # Save results as CSV
        self.sig_worker_rescsv.connect(self.worker.save_results_csv)
        self.worker.sig_rescsv_finished.connect(self.save_results_csv_finished)

        # Load from CSV
        self.sig_worker_loadcsv.connect(self.worker.load_csv_file)
        self.worker.sig_load_csv_finished.connect(self.load_csv_file_finished)

    def connect_parameter_tree(self) -> None:
        """Connect signals from the Configuration widget."""
        self.wconfiguration.sig_syncroi_changed.connect(self.syncroi_changed)

    def run_after_tree_change(
        self, context: str, param_name: str, value: float | int | list | bool | str
    ) -> None:
        """Run additional actions after a parameter change in the parameter tree."""
        if context != "settings":
            return

        match param_name:
            case "poly_window":
                if isinstance(value, list):
                    self.update_roi_from_poly(value)
            case "fft_window":
                if isinstance(value, list):
                    self.update_roi_from_fft(value)
            case "max_time":
                self.crop_signal()
            case "time_offset":
                self.apply_time_offset()
            case "offset":
                self.plot_tdo_detrended()
            case param_name if param_name in PARAM_ANALYSIS_TRIGGER:
                self.run_analysis()

    @pyqtSlot()
    def load_file(self):
        """
        Load a configuration or NeXus file and start a worker.

        Callback for when the "File" parameter changes, or a file is drag & dropped in
        the window.
        """
        super().load_file()

        # Set the ROIs
        # Turn off analysis because we need to extract the TDO signal first
        self._do_analysis = False
        region1 = self.worker.proc.cfg.settings.poly_window
        region2 = self.worker.proc.cfg.settings.fft_window
        if isinstance(region1, Sequence) and len(region1) == 2:
            self.update_roi_from_poly(region1)
        if isinstance(region2, Sequence) and len(region2) == 2:
            self.update_roi_from_fft(region2)

        if self.worker.is_data_loaded:
            # If data was loaded automatically, we can show it
            self.load_data_finished()

        if self.worker.proc.is_nexus_file:
            # If NeXus file, data and results can be shown
            self.load_data_finished()
            self.preprocess_finished()
            self.run_analysis_finished()

        self._do_analysis = True  # turn analysis back on

    @pyqtSlot()
    def crop_signal(self):
        """Crop the signal."""
        self.disable_buttons()
        self.sig_worker_crop.emit()

    def crop_signal_finished(self):
        """Re-enable buttons and print a message."""
        self.enable_buttons()
        self.logger.info(
            "[GUI] Signal was cropped, please re-run the TDO signal extraction"
        )

    @pyqtSlot()
    def preprocess_finished(self):
        """Plot the TDO signal."""
        # Clear plots since source signal was changed
        self.wgraphs.tdo_field.clearPlots()
        self.wgraphs.tdo_inverse_field.clearPlots()
        self.wgraphs.fft.clearPlots()

        self.align_field_finished()
        self.plot_tdo()
        self.enable_buttons()

    def apply_time_offset(self):
        """Apply an offset between the pickup and the signal time bases."""
        if not self.check_data_loaded():
            return
        self.disable_buttons()
        self.sig_worker_offset.emit()

    @pyqtSlot()
    def time_offset_finished(self):
        """Re-plot signals after applying the offset."""
        self.align_field_finished()
        self.plot_tdo()
        self.run_analysis()

    @pyqtSlot()
    def run_analysis(self):
        """Fit, detrend, oversample in 1/B and compute the FFT."""
        if not self.check_tdo_extracted():
            self.logger.warning("[GUI] TDO signal was not extracted")
            return
        
        super().run_analysis()

    @pyqtSlot()
    def run_analysis_finished(self):
        """Plot the detrended signals and the FFT after analysis."""
        self.plot_tdo_detrended()
        self.plot_fft()
        self.enable_buttons()

    @pyqtSlot()
    def save_tdo_csv(self):
        """Save extracted TDO signal as CSV."""
        if not self.check_data_loaded():
            return

        self.disable_buttons()

        default_fname = self.worker.proc.get_csv_filename(suffix="-tdo")
        fname, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Save TDO signals as...",
            default_fname,
            "Text files (*.txt, *.csv, *.tsv, *.out)",
        )

        if fname:
            self.sig_worker_tdocsv.emit(fname)
        else:
            self.logger.error("[GUI] Invalid output file name for CSV file")
            self.save_tdo_csv_finished()

    @pyqtSlot()
    def save_tdo_csv_finished(self):
        """Re-enable buttons."""
        self.enable_buttons()

    @pyqtSlot()
    def save_results_csv(self):
        """Save final results as CSV."""
        if not self.check_data_loaded():
            return

        self.disable_buttons()

        default_fname = self.worker.proc.get_csv_filename(suffix="-results")
        fname, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Save TDO signals as...",
            default_fname,
            "Text files (*.txt, *.csv, *.tsv)",
        )

        if fname:
            self.sig_worker_rescsv.emit(fname)
        else:
            self.logger.error("[GUI] Invalid output file name for CSV file")
            self.save_tdo_csv_finished()

    @pyqtSlot()
    def save_results_csv_finished(self):
        """Re-enable buttons."""
        self.enable_buttons()

    def load_csv_file(self, file_path: str):
        """Load a CSV file."""
        if not self.check_config_loaded():
            self.logger.error(
                "[GUI] Can't load a CSV file without loading a configuration file first"
            )
            return

        self.disable_buttons()
        self.sig_worker_loadcsv.emit(file_path)

    @pyqtSlot()
    def load_csv_file_finished(self):
        """Trigger action as if the TDO signal was extracted (but it was loaded)."""
        self.preprocess_finished()
        self.run_analysis_finished()

    @pyqtSlot()
    def batch_process(self):
        """Required for compatibility with the base app, but it is not implemented."""
        self.logger.warning("[GUI] Not implemented")

    @pyqtSlot()
    def handle_roi1_changed(self):
        """
        Define what to do when the ROI in the TDO panel is changed.

        The "poly_window" setting is updated in the tree, a new analysis is requested.
        """
        self._handle_roi_changed(self.wgraphs.roi, "poly_window")

        if self.check_tdo_extracted():
            self.run_analysis()

    @pyqtSlot()
    def handle_roi2_changed(self):
        """
        Define what to do when the ROI in the TDO panel is changed.

        The "fft_window" setting is updated in the tree, a new analysis is requested.
        """
        self._handle_roi_changed(self.wgraphs.roi2, "fft_window")

        if self.check_tdo_extracted():
            self.run_analysis()

    def update_roi_from_poly(self, new_region: Sequence[float | int]):
        """Update the ROI in the graph from "Fit: field window" in the parameter tree."""
        self._update_roi(self.wgraphs.roi, new_region)

    def update_roi_from_fft(self, new_region: Sequence[float | int]):
        """Update the ROI in the graph from "FFT: field window" in the parameter tree."""
        self._update_roi(self.wgraphs.roi2, new_region)

    def plot_tdo(self):
        """Plot TDO signal versus field and time."""
        if not self.check_field_aligned():
            self.worker.align_field()

        ## TDO signal versus field
        self.wgraphs.sig_field.clearPlots()

        # Decreasing magnetic field
        self.wgraphs.sig_field.plot(
            self.worker.proc.get_data_processed("magfield")[self.ind_bdown],
            self.worker.proc.get_data_processed(self.worker.proc._tdo_name)[
                self.ind_bdown
            ],
            pen=self.wgraphs.pen_bdown,
            name="B down",
        )
        # Increasing magnetic field
        self.wgraphs.sig_field.plot(
            self.worker.proc.get_data_processed("magfield")[self.ind_bup],
            self.worker.proc.get_data_processed(self.worker.proc._tdo_name)[
                self.ind_bup
            ],
            pen=self.wgraphs.pen_bup,
            name="B up",
        )

        ## TDO signal versus time
        self.wgraphs.sig_time.clearPlots()

        # Decreasing magnetic field
        self.wgraphs.sig_time.plot(
            self.worker.proc.get_data_processed("time_exp")[self.ind_bdown],
            self.worker.proc.get_data_processed(self.worker.proc._tdo_name)[
                self.ind_bdown
            ],
            pen=self.wgraphs.pen_bdown,
            name="B down",
        )
        # Increasing magnetic field
        self.wgraphs.sig_time.plot(
            self.worker.proc.get_data_processed("time_exp")[self.ind_bup],
            self.worker.proc.get_data_processed(self.worker.proc._tdo_name)[
                self.ind_bup
            ],
            pen=self.wgraphs.pen_bup,
            name="B up",
        )

    @pyqtSlot()
    def plot_tdo_detrended(self):
        """Plot TDO signal with background removed."""
        if not self.check_tdo_detrended():
            return
        # Get parameters
        offset = self.worker.proc.cfg.settings.offset
        lower_b, upper_b = self.worker.proc.cfg.settings.fft_window

        # Sync fit and FFT windows
        if lower_b == -1:
            lower_b = self.worker.proc.cfg.settings.poly_window[0]
        if upper_b == -1:
            upper_b = self.worker.proc.cfg.settings.poly_window[1]

        ## TDO detrended versus field
        self.wgraphs.tdo_field.clearPlots()
        self.wgraphs.sig_field.clearPlots()
        self.plot_tdo()  # to make sure data shown is synced

        # Decreasing magnetic field
        if self.worker.proc.get_data_processed(
            self.worker.proc._tdo_det_dec_name, checkonly=True
        ):
            self.wgraphs.tdo_field.plot(
                self.worker.proc.get_data_processed("magfield")[self.ind_bdown],
                self.worker.proc.get_data_processed(self.worker.proc._tdo_det_dec_name)
                - offset / 2,
                pen=self.wgraphs.pen_bdown,
                name="B down",
            )
        # Increasing magnetic field
        if self.worker.proc.get_data_processed(
            self.worker.proc._tdo_det_inc_name, checkonly=True
        ):
            self.wgraphs.tdo_field.plot(
                self.worker.proc.get_data_processed("magfield")[self.ind_bup],
                self.worker.proc.get_data_processed(self.worker.proc._tdo_det_inc_name)
                + offset / 2,
                pen=self.wgraphs.pen_bup,
                name="B up",
            )

        # Fit decreasing magnetic field
        if self.worker.proc.get_data_processed(
            self.worker.proc._tdo_name + "_fit_dec", checkonly=True
        ):
            self.wgraphs.sig_field.plot(
                self.worker.proc.get_data_processed("magfield")[self.ind_bdown],
                self.worker.proc.get_data_processed(
                    self.worker.proc._tdo_name + "_fit_dec"
                ),
                pen=self.wgraphs.pen_fitdown,
                name="Fit B down",
            )
        # Fit increasing magnetic field
        if self.worker.proc.get_data_processed(
            self.worker.proc._tdo_name + "_fit_inc", checkonly=True
        ):
            self.wgraphs.sig_field.plot(
                self.worker.proc.get_data_processed("magfield")[self.ind_bup],
                self.worker.proc.get_data_processed(
                    self.worker.proc._tdo_name + "_fit_inc"
                ),
                pen=self.wgraphs.pen_fitbup,
                name="Fit B up",
            )

        ## TDO detrended versus 1/B
        self.wgraphs.tdo_inverse_field.clearPlots()

        # Plot the whole non-interpolated signal
        # Increasing magnetic field
        if self.worker.proc.get_data_processed(
            self.worker.proc._tdo_det_inc_name, checkonly=True
        ):
            self.wgraphs.tdo_inverse_field.plot(
                1 / self.worker.proc.get_data_processed("magfield")[self.ind_bup],
                self.worker.proc.get_data_processed(self.worker.proc._tdo_det_inc_name)
                + offset / 2,
                pen=self.wgraphs.pen_tdo,
            )
        # Decreasing magnetic field
        if self.worker.proc.get_data_processed(
            self.worker.proc._tdo_det_dec_name, checkonly=True
        ):
            self.wgraphs.tdo_inverse_field.plot(
                1 / self.worker.proc.get_data_processed("magfield")[self.ind_bdown],
                self.worker.proc.get_data_processed(self.worker.proc._tdo_det_dec_name)
                - offset / 2,
                pen=self.wgraphs.pen_tdo,
            )

        # Now, the oversampled signal
        # Decreasing magnetic field
        if self.worker.proc.get_data_processed(
            self.worker.proc._tdo_inv_dec_name, checkonly=True
        ):
            p0 = self.wgraphs.tdo_inverse_field.plot(
                self.worker.proc.get_data_processed("magfield_inverse_dec"),
                self.worker.proc.get_data_processed(self.worker.proc._tdo_inv_dec_name)
                - offset / 2,
                pen=self.wgraphs.pen_bdown,
                name="B down",
            )
        # Increasing magnetic field
        if self.worker.proc.get_data_processed(
            self.worker.proc._tdo_inv_inc_name, checkonly=True
        ):
            p1 = self.wgraphs.tdo_inverse_field.plot(
                self.worker.proc.get_data_processed("magfield_inverse_inc"),
                self.worker.proc.get_data_processed(self.worker.proc._tdo_inv_inc_name)
                + offset / 2,
                pen=self.wgraphs.pen_bup,
                name="B up",
            )
        # Adjust limits
        self.wgraphs.tdo_inverse_field.getViewBox().autoRange(
            padding=0.1, items=[p0, p1]
        )

    def plot_fft(self):
        """Plot FFT in 1/B."""
        self.wgraphs.fft.clearPlots()

        # Decreasing magnetic field
        if self.worker.proc.get_data_processed("fft_dec", checkonly=True):
            self.wgraphs.fft.plot(
                self.worker.proc.get_data_processed("bfreq_dec"),
                self.worker.proc.get_data_processed("fft_dec"),
                pen=self.wgraphs.pen_bdown,
                name="B down",
            )
        # Increasing magnetic field
        if self.worker.proc.get_data_processed("fft_inc", checkonly=True):
            self.wgraphs.fft.plot(
                self.worker.proc.get_data_processed("bfreq_inc"),
                self.worker.proc.get_data_processed("fft_inc"),
                pen=self.wgraphs.pen_bup,
                name="B up",
            )

    @pyqtSlot()
    def syncroi_changed(self):
        """Update fit and FFT windows sync. status."""
        self.wgraphs._sync_roi = self.wconfiguration.syncroi_parameter.value()

    @pyqtSlot(bool, str)
    def select_file_in_browser(self, is_toml: bool, filepath: str):
        """
        Select a file and set it as a configuration file or a new experiment ID.

        Callback for when the user double-click on a file in the "Files" tab.
        """
        if filepath.endswith(".csv"):
            self.load_csv_file(filepath)
        else:
            super().select_file_in_browser(is_toml, filepath)

    def check_tdo_extracted(self) -> bool:
        """Check if the TDO signal was extracted."""
        if not self.check_data_loaded():
            return False
        return self.worker.proc._check_barycenters_computed()

    def check_tdo_detrended(self) -> bool:
        """Check if the detrending was performed."""
        if not self.check_tdo_extracted():
            return False
        return self.worker.proc._check_tdo_detrended()

    def dropEvent(self, a0: QtGui.QDropEvent | None = None) -> None:
        """Load a file when it is dropped in the main window."""
        if a0 is None:
            return
        mime_data = a0.mimeData()
        if mime_data is None:
            return
        for url in mime_data.urls():
            file_path = url.toLocalFile()
            if file_path.endswith(OUT_FORMAT):
                self.load_csv_file(file_path)
                return

        super().dropEvent(a0)



"""Config class for the LinxProcessor object."""

from pathlib import Path
from typing import Any

from ..core import BaseConfig
from .config_models import LinxConfigurationModel

CONFIG_DEFAULT = Path(__file__).parent / "assets" / "config_default.toml"


class LinxConfig(BaseConfig):
    def __init__(
        self,
        user_file: str | Path | None = None,
        default_file: str | Path | None = None,
        **overrides: Any,
    ) -> None:
        if not default_file:
            default_file = CONFIG_DEFAULT

        super().__init__(LinxConfigurationModel, user_file, default_file, **overrides)

    def load(self, merge_measurement_dict: bool = True) -> None:
        """Load."""
        super().load(merge_measurement_dict=merge_measurement_dict)

    @classmethod
    def loads(cls, json_data: str | bytes, **kwargs) -> "LinxConfig":
        return super()._loads(LinxConfigurationModel, json_data, **kwargs)



"""Qt base widgets to re-use and customise."""

from ._batch_processing import BaseBatchProcessingWidget
from ._configuration import BaseConfigurationWidget
from ._files import FileBrowserWidget
from ._graphs import BaseGraphsWidget
from ._param_content import BaseParamContent
from ._popup_progressbar import PopupProgressBar
from ._text_logger import TextLoggerWidget

__all__ = [
    "BaseConfigurationWidget",
    "BaseGraphsWidget",
    "BaseBatchProcessingWidget",
    "BaseParamContent",
    "FileBrowserWidget",
    "PopupProgressBar",
    "TextLoggerWidget",
]



"""The EchoProcessor class for ultrasound echoes experiments."""

import logging
import os
import time
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Any, Literal, Self

import nexusformat.nexus as nx
import numpy as np
from fftekwfm import TekWFM
from scipy import signal

from ..core import BaseProcessor, sp
from ..log import configure_logger
from ._config import EchoConfig

# Name of the analysis serie, here it correspond to an echo index
SERIE_NAME = "echo"
# Measurements names to use
REFNAME = "reference"
I_NAME_ANALOG = "in_phase"
Q_NAME_ANALOG = "out_phase"
I_NAME_DIGITAL = "in_phase_demod"
Q_NAME_DIGITAL = "out_phase_demod"
# To guess if time is in microseconds, set approx. duration of a frame in seconds
# If the max is found 1000x above, it is guessed that the vector is in microseconds
FRAME_TIMESCALE = 10e-6

logger = logging.getLogger(__name__)


class EchoProcessor(BaseProcessor):
    """Processor class for ultrasound echoes experiments."""

    cfg: EchoConfig

    def __init__(self, *args, digital: bool | None = None, **kwargs) -> None:
        """
        Processor class for ultrasound echoes experiments.

        Provide methods for :
        - Loading data and metadata,
        - Reshape and format data,
        - Preprocessing : demodulation, decimation and/or smoothing (with rolling window
            average)
        - Averaging in a time-window,
        - Derive ultrasound waves attenuation and phase-shift in the sample during a
            magnetic field shot.

        The data can be loaded from both binary files from the LabVIEW program or
        Tektronix WFM files. Digital demodulation is possible when necessary.

        In any case, the raw data is assumed to be in the form of several frames paving
        the experiment duration. Each frame should contain the same number of samples.
        This is all read from the inputs files :

        With the original LabVIEW program :
        - sample.bin : raw binary file from the LabVIEW program with the measured
            channels,
        - sample.txt : frames onsets, with metadata in the header. The header is
            specified in the configuration file,
        - sample-pickup.bin : raw binary file from the LabVIEW program with pickup coil
            voltage.

        With the Tektronix oscilloscope :
        - sample_ch1.wfm, sample_ch2.wfm, ... : Tektronix WFM files with the measured
            channels,
        - sample.txt : metadata only, the frames onsets are read from the WFM file,
        - sample-pickup.bin : see above.

        Parameters
        ----------
        *args : passed to `BaseProcessor()`
        digital : bool or None, optional
            Force digital mode by setting this parameter to True or to analog mode with
            False. None will attempt to determine this automatically from the
            configuration file, this is the default behavior.
        **kwargs : passed to `BaseProcessor()`
        """
        configure_logger(logger, "pyuson.log")

        # Get the Config object
        self._config_cls = EchoConfig

        # Prepare internals names
        self._serie_name = SERIE_NAME

        # Use digital mode or not, or automatically detect it from configuration
        self.is_digital = digital

        # Convenience attributes (might be overidden upon loading)
        self.nframes = 0
        self.nsamples = 0

        super().__init__(*args, **kwargs)

        # Initialize flags
        if not self.is_nexus_file:
            self._init_flags()

            # Additional initializations
            if self.is_digital:
                self._init_digital()
            else:
                self._i_name = I_NAME_ANALOG
                self._q_name = Q_NAME_ANALOG

            # Initialize metadata dict for WFM file (will be filled when loading data)
            self.metadata["vscale"] = dict()
            self.metadata["voffset"] = dict()

    @property
    def idx_serie(self) -> int:
        """
        Track the current analyzed echo.

        The returned value is read from the "echo_index" parameter in the "settings"
        section of the configuration.
        """
        return self.cfg.settings.echo_index

    @idx_serie.setter
    def idx_serie(self, value: int):
        """
        Setter for `idx_serie`.

        The "echo_index" parameter in the "settings" section of the configuration is
        updated.
        """
        self.cfg.settings.echo_index = value

    @property
    def analysis_window(self) -> Sequence[float]:
        """
        Time-range in which averaging is performed.

        The returned value is read from the "analysis_window" parameter in the
        "settings" section of the configuration.
        """
        return self.cfg.settings.analysis_window

    @analysis_window.setter
    def analysis_window(self, value: Sequence[float]):
        """
        Setter for `idx_serie`.

        The "analysis_window" parameter in the "settings" section of the configuration
        is updated.
        """
        if len(value) != 2:
            logger.error(f"Invalid analysis window : {value}")
        if value != self.cfg.settings.analysis_window:
            logger.info(f"Analysis window set to {value[0], value[1]}")
            self.cfg.settings.analysis_window = value

    def _init_digital(self) -> None:
        """Additional initialization for digital mode."""
        # Prepare internal names
        self._ref_name = REFNAME
        self._i_name = I_NAME_DIGITAL
        self._q_name = Q_NAME_DIGITAL
        # Get signal name
        if self.is_config_file:
            # If instantiated with configuration file
            self._sig_name = self._get_signal_name()
        else:
            self._sig_name = "SIG_NAME_NOT_SET"  # might be set later

        # Update flags
        self.is_digital = True

        # Add a seed based on unix time for frame selection
        self._seed = int(time.time())

        # Update measurements names, eg. the actual measurements used in computation and
        # not orignal "signal" and "reference"
        self.measurements = [self._i_name, self._q_name]
        # Update flags
        self.is_rollmean = {m: False for m in self.measurements}  # the keys changed

    def _reinit(self) -> None:
        """Re-initialize data objects."""
        super()._reinit()
        if self.is_digital:
            self._init_digital()

        self.metadata["vscale"] = dict()
        self.metadata["voffset"] = dict()
        self.nframes = 0
        self.nsamples = 0

    def load_pickup(self) -> Self:
        """Load pickup coil binary data and create the corresponding time vector."""
        # Get parameters
        filename = self.cfg.filenames["pickup"]
        precision = self.cfg.files["pickup"].precision
        endian = self.cfg.files["pickup"].endian
        order = self.cfg.files["pickup"].order
        npickups = self.cfg.parameters.pickup_number
        pickup_index = self.cfg.parameters.pickup_index
        samplerate = self.cfg.parameters.pickup_samplerate

        # Read and store pickup coil voltage
        logger.info(f"Loading pickup data from from {filename.name}...")
        if Path(filename).is_file():
            self.set_data_raw(
                "pickup",
                self._load_pickup(
                    filename,
                    precision,
                    endian,
                    order=order,
                    nseries=npickups,
                    index=pickup_index,
                ),
            )
            # Create and store corresponding time vector
            nsamples = self.data_raw["pickup"].shape[0]
            self.set_data_raw(
                "pickup_time",
                np.linspace(0, nsamples / samplerate, nsamples, endpoint=False),
            )

            # Verbose
            logger.info("Pickup signal loaded")
        else:
            logger.info(f"{filename.name} not found, skipping")
            # Create fake data that will be created when we'll know how long is the
            # experiment
            self.set_data_raw("pickup", np.array([]))
            self.set_data_raw("pickup_time", np.array([]))

        return self

    def load_data(self, **kwargs) -> Self:
        """Load the raw data."""
        self.load_oscillo(**kwargs)
        return self

    def preprocess(self, **kwargs) -> Self:
        """Run the preprocessing steps before analysis, i.e. run the demodulation."""
        if self.is_digital:
            self.demodulate(**kwargs)
        return self

    def align_field(self, on_pickup: bool = False, **kwargs) -> Self:
        """
        Align magnetic field time serie on the experiment time vector.

        The pickup-coil sampling rate is much higher than the frames sampling rate. In
        order to plot the different resulting metrics against the magnetic field instead
        of the time, it needs to be aligned on the same time base using linear
        interpolation.

        Note that the magnetic field serie is overwritten with the aligned version.

        **kwargs are ignored.
        """
        # Checks
        if not self.get_data_processed("time_exp", checkonly=True):
            self.get_time_exp()
            # maybe it was still not computed, we need to recheck
            if not self.get_data_processed("time_exp", checkonly=True):
                logger.error("Time vector could not be built, exiting")
                return self

        # We need to re-compute the field from scratch in case it was downsampled with
        # rolling average to make sure we're not upscaling it. This could be smarter (
        # by tracking the switches between rolling-average/no-rolling-average) but the
        # computation is cheap
        self.compute_field()

        super().align_field(on_pickup=on_pickup, left=0, right=0)

        return self

    def run_analysis(self) -> Self:
        """
        Helper function that runs the analysis.

        Average the frames, compute attenuation and phase-shift.
        """
        self.average_frame_range().compute_attenuation().compute_phase_shift()
        return self

    def load(self, filename: str | Path | None) -> Self:
        """
        Load a previously created NeXus file.

        Use `BaseProcessor.load()` method, with additionnal steps : guess the internal
        flags and add convenience attributes. The returned `EchoProcessor` object can be
        used to resume analysis where it was left off.

        Parameters
        ----------
        filename : str
            Full path to the file to load.
        """
        super().load(filename)

        # Set measurements names
        self.measurements = [*self.cfg.measurements.keys()]

        # Update flags
        self._guess_flags()

        # Update convenience attributes
        if not self.is_digital:
            self._i_name = I_NAME_ANALOG
            self._q_name = Q_NAME_ANALOG
        if "frame_onsets" in self.data_raw:
            self.nframes = self.data_raw["frame_onsets"].shape[0]
        else:
            logger.warning("Could not read number of frames, features will be limited.")
        if "time_meas" in self.data_raw:
            self.nsamples = self.data_raw["time_meas"].shape[0]
        else:
            logger.warning(
                "Could not read number of measurement time points, features will be"
                "limited."
            )

        return self

    def export_csv(
        self, fname: None | str | Path = None, sep: str = "\t", to_cm: bool = False
    ) -> None:
        r"""
        Export attenuation and phase shift data as a CSV file.

        Columns names are written on the first line. The current echo index is used to
        select which data to save.

        Parameters
        ----------
        fname : str, Path or None, optional
            Full path to the output file. If None, the name is generated automatically.
            This is the default behavior.
        sep : str, optional
            Separator in the file, usually "," (csv) or "\t" (tsv, default).
        to_cm : bool, optional
            Convert attenuation from dB/m to dB/cm. Default is False.
        """
        # Create output file name
        if not fname:
            fname = self.get_csv_filename()

        logger.info(f"Saving at {fname}...")

        # Collect data to be saved
        attenuation = self.get_data_serie("attenuation")[..., np.newaxis]
        data = np.concatenate(
            (
                self.get_data_processed("time_exp")[..., np.newaxis],
                self.get_data_processed("magfield")[..., np.newaxis],
                self.get_data_serie("phase_avg")[..., np.newaxis],
                self.get_data_serie("amplitude_avg")[..., np.newaxis],
                attenuation / 100 if to_cm else attenuation,
                self.get_data_serie("phaseshift")[..., np.newaxis],
            ),
            axis=1,
        )
        # Configure output file
        header = sep.join(
            ("time", "field", "phase", "amplitude", "attenuation", "dphase")
        )

        # Save CSV file
        np.savetxt(fname, data, delimiter=sep, header=header, comments="")

        logger.info("Done.")

    def rolling_average(self) -> Self:
        """
        Perform a rolling average on raw data measurements.

        Store the smoothed arrays in `data_processed[results]` with a leading "s".

        The data can optionnally be sub-sampled, given the parameters defined in the
        settings section of the configuration file, in that case the corresponding time
        vector is subsampled as well and store in `data_processed[results]`.

        Uses the `sp.rolling_average()` function.
        """
        # Collect parameters
        wlen = self.cfg.settings.rolling_mean_wlen
        subsample = self.cfg.settings.rolling_mean_subsample

        # Check if there is something to do
        if wlen < 2:
            # wlen = 1 or 0 means no rolling average
            logger.warning(f"Time window set to {wlen}, no rolling average")

            # Re-generate time vector in case subsampling was applied before
            self.get_time_exp()
            self.align_field()

            # Set flags
            self.is_rollmean.update({m: False for m in self.measurements})
            return self

        for measurement in self.measurements:
            # Get data (do not use the dedicated method to be sure to start from raw
            # data and not already smoothed trace)
            data = (
                self.get_data_processed(measurement)
                if self.is_digital
                else self.get_data_raw(measurement)
            )

            # Check data is 2D
            if not self._check_frames_2d(measurement):
                logger.error(
                    f"{measurement} was not reshaped, can't apply rolling average"
                )
                return self

            logger.info(
                f"Applying rolling average to {measurement} with a window of {wlen} "
                "samples..."
            )

            self.set_data_processed(
                f"s{measurement}",
                sp.rolling_average(data, wlen, subsample=subsample, axis=1),
            )
            self.is_rollmean[measurement] = True
            logger.info(f"{measurement} was smoothed")

        # Regenerate the experiment time vector in case it was subsampled before
        self.get_time_exp()

        # If subsampling is on, we need to update the experiment time vector
        if subsample:
            logger.info("Subsampling is on, updating time and magnetic field...")
            self.set_data_processed(
                "time_exp",
                sp.subsample_array(self.get_data_processed("time_exp"), wlen - 1),
            )

        # Re-align magnetic field
        self.align_field()

        return self

    def batch_process(
        self,
        expids: Sequence,
        rolling_average: bool = False,
        save_csv: bool = False,
        save_csv_kwargs: dict = {},
        force_find_f0: bool = False,
        *,
        demodulation_progress_emitter: Any = None,
        batch_progress_emitter: Any = None,
    ) -> Self:
        """
        Batch-process a list of experiment IDs, keeping current parameters.

        Only supports datasets present in the current `data_directory`. Optionally,
        export the results as a CSV file for each dataset.

        Parameters
        ----------
        expids : Sequence
            List of experiment ID to process.
        rolling_average : bool, optional
            Whether to apply rolling average in the process. Default is False.
        save_csv : bool, optional
            Whether to export results as a CSV file. Default is False.
        save_csv_kwargs : dict, optional
            Used only when `save_csv` is True. Specify arguments for the `to_csv()`
            method. Default is an empty dict (default arguments will be used).
        force_find_f0 : bool, optional
            Force finding f0 in digital mode, whatever the value of `f0` in the
            demodulation section of the current configuration. Default is False.
        demodulation_progress_emitter : Any, optional
            An object with an `emit()` method, such as a pyqtSignal. The loop index is
            emitted at each iteration of the demodulation loop. Default is None.
        batch_progress_emitter : Any, optional
            An object with an `emit()` method, such as a pyqtSignal. The loop index is
            emitted at each iteration of the main loop. Default is None.
        """
        for idx, expid in enumerate(expids):
            # Set the experiment ID (will reinitialize data but not configuration)
            self.expid = expid

            # Load data
            self.load_oscillo(scale=True)
            # Load and align magnetic field
            self.align_field()

            # Demodulate
            if self.is_digital:
                if force_find_f0:
                    self.find_f0()
                self.demodulate(progress_emitter=demodulation_progress_emitter)

            # Rolling average
            if rolling_average:
                self.rolling_average()

            # Average frames and compute
            self.average_frame_range().compute_attenuation().compute_phase_shift()

            # Export as CSV
            if save_csv:
                self.export_csv(**save_csv_kwargs)

            # Emit progress
            if batch_progress_emitter is not None:
                batch_progress_emitter.emit(idx)

        return self

    def load_metadata(self) -> Self:
        """
        Read metadata from the header of the text file with the frames onsets.

        Wrap the `_load_metadata()` method, fetching the parameters from the
        configuration.
        """
        # Get parameters
        filename = self.cfg.filenames["reference_time"]
        header_map = self.cfg.metadata.index_map
        conversion_map = self.cfg.metadata.conversion_map

        # Read metadata
        logger.info(f"Reading metadata from {os.path.basename(filename)}...")
        metadata = self._load_metadata(filename, header_map, conversion_map)

        # Remove unused metadata for WFM files
        if self.is_wfm:
            metadata.pop("dt_acq", None)  # will be read from the WFM file directly

        # Store (add freshly loaded metadata)
        self.metadata = self.metadata | metadata

        if "rf_frequency" in self.metadata:
            # Add RF frequency offset
            offset = self.cfg.parameters.rf_offset
            self.metadata["rf_frequency"] = self.metadata["rf_frequency"] + offset
            logger.info(
                f"RF: {self.metadata['rf_frequency'] / 1e6:3.3f}MHz read from "
                f"file (including a {offset / 1e6:3.3f}MHz offset)"
            )
        else:
            logger.error("Could not read RF frequency from metadata file")

        return self

    def load_frame_onsets(self) -> Self:
        """
        Read frames onsets in the reference time text file.

        For LabVIEW binary files only. For WFM files, frame onsets are read from the
        file.

        Wrap the `_load_frame_onsets()` method, fetching the parameters from the
        configuration.
        """
        if self.is_wfm:
            # This will be read from the WFM file header.
            return self

        # Get parameters
        filename = self.cfg.filenames["reference_time"]
        nlines_header = self.cfg.files["reference_time"].header
        delimiter = self.cfg.files["reference_time"].delimiter

        # Read and store frames onsets
        if not filename.is_file():
            logger.error(
                f"{filename.name} does not exist, check your configuration file"
            )
            return self

        logger.info(f"Reading frame onsets from {filename.name}...")
        self.set_data_raw(
            "frame_onsets", self._load_frame_onsets(filename, nlines_header, delimiter)
        )
        self.nframes = self.data_raw["frame_onsets"].shape[0]

        # Verbose
        logger.info("Frame onsets loaded")

        return self

    def load_oscillo(self, **kwargs) -> Self:
        """
        Load oscilloscope data.

        Arguments are passed to the loader method, either `load_oscillo_wfm()` for WFM
        files, or `load_oscillo_bin()` for LabVIEW binary files.

        The frames are reshaped if needed and the time vector for both the measurement
        and the experiment are built and stored in `data_raw[results]`.

        Parameters
        ----------
        **kwargs : passed to `load_oscillo_wfm()`.
        """
        filename = self.cfg.filenames["oscillo"]

        # Attempt to load metadata first
        try:
            self.load_metadata()
        except Exception as e:
            logger.warning(f"Metadata could not be loaded : {e}")
            pass

        if filename.suffix.endswith("wfm"):
            self.is_wfm = True
            self.load_oscillo_wfm(**kwargs)
            self.metadata["toffset_meas"] = self.get_data_raw("time_meas")[0]
        elif filename.suffix.endswith("bin"):
            self.is_wfm = False
            self.load_oscillo_bin()
            self.reshape_frames()
        else:
            logger.warning(
                f"{filename.name} : extension not recognized, assuming LabVIEW binary "
                "file."
            )
            self.is_wfm = False
            self.load_oscillo_bin()

        # Load experiment time vector
        self.get_time_exp()
        
        if (self.nframes > 0) and (self.nsamples > 0):
            logger.info(f"Data loaded : {self.nframes} frames, {self.nsamples} samples")
        else:
            raise ValueError("Error loading file")

        return self

    def load_oscillo_bin(self) -> Self:
        """Load oscilloscope binary data from LabVIEW binary files."""
        # Get parameters
        filename = self.cfg.filenames["oscillo"]
        nchannels = len(self.cfg.measurements)
        precision = self.cfg.files["oscillo"].precision
        endian = self.cfg.files["oscillo"].endian
        order = self.cfg.files["oscillo"].order

        # Check file exists
        if not filename.is_file():
            logger.error(
                f"{filename.name} does not exist, check your configuration file"
            )
            return self

        # Read data
        logger.info(f"Loading oscilloscope data from {filename.name}...")
        res = self._load_oscillo_bin(filename, nchannels, precision, endian, order)

        # Store
        for meas_name, meas_index in self.cfg.measurements.items():
            self.set_data_raw(meas_name, res[meas_index])
            # Add rolling average flag
            self.is_rollmean[meas_name] = False
            # Add dummy voltage scaling for compatibility with WFM files
            self.metadata["vscale"][meas_name] = 1
            self.metadata["voffset"][meas_name] = 0

        # Verbose
        logger.info(f"{', '.join(x for x in self.cfg.measurements.keys())} loaded")

        return self

    def load_oscillo_wfm(self, scale: bool = False, microseconds: bool = True) -> Self:
        """
        Load oscilloscope data from Tektronix WFM files.

        Parameters
        ----------
        scale : bool, optional
            Whether to rescale loaded data to real units (e.g. volts) or keep it in
            int16. Default is False.
        microseconds : bool, optional
            Whether to convert the measurement time vector `time_meas`, to microseconds.
            Default is True.
        """
        # Get parameters
        filename_base = self.cfg.filenames["oscillo"]

        for meas_name, meas_index in self.cfg.measurements.items():
            # Replace the channel placeholder with the actual channel index
            filename = Path(
                str(filename_base).replace("_!CHANNELID", f"_ch{meas_index}")
            )

            if not filename.is_file():
                # try uppercase extension
                filename = filename.with_suffix(".WFM")
                if not filename.is_file():
                    logger.error(
                        f"{filename.name} doesn't exist, check your configuration file"
                    )
                    return self

            logger.info(f"Loading oscilloscope data from {filename.name}...")
            # Load data
            tek = TekWFM(filename).load_frames().get_time_frame()

            # Store data and metadata
            if scale:
                self.set_data_raw(meas_name, tek.scale_data(tek.frames))
            else:
                self.set_data_raw(meas_name, tek.frames)

            self.metadata["vscale"][meas_name] = tek.vscale
            self.metadata["voffset"][meas_name] = tek.voffset

            logger.info(f"{meas_name} loaded")

        # Store (meta)data
        self.set_data_raw("time_meas", tek.time_frame)
        if microseconds:
            self.is_us = True
            self.data_raw["time_meas"] *= 1e6
        else:
            self.is_us = False
        self.set_data_raw("frame_onsets", tek.frame_onsets)
        self.metadata["dt_acq"] = tek.tscale
        self.metadata["frame_duration"] = tek.npoints * tek.tscale
        self.nframes = tek.nframes
        self.nsamples = tek.npoints

        return self

    def reshape_frames(
        self, microseconds: bool = True, meas_names: None | str | Iterable[str] = None
    ) -> Self:
        """
        Reshape oscillo channels array to get a column per frame.

        Each measurement yields a 2D array with time series on columns, thus an array
        with sahep (npoints, nframes).

        Also build the corresponding measurement time vector.

        Parameters
        ----------
        microseconds : bool, optional
            Convert the time vector from seconds to microseconds. Default is True.
        meas_names : None or Sequence, optional
            If None (default), uses all available measurement names.

        """
        # Checks
        if self.is_wfm:
            return self

        if "frame_onsets" not in self.data_raw:
            self.load_frame_onsets()

        # Get list of arrays to reshape
        if not meas_names:
            measurements = self.measurements
        elif isinstance(meas_names, str):
            measurements = [meas_names]
        else:
            measurements = meas_names

        # Reshape frames in-place
        reshaped = False
        for measurement in measurements:
            logger.debug(f"Reshaping {measurement}...\t")
            # check if it was loaded
            if measurement not in self.data_raw:
                logger.warning(f"{measurement} was not loaded, skipping")
                continue

            if self._check_frames_2d(measurement):
                # check if it makes sense to reshape the data
                logger.warning(
                    f"{measurement} is not 1D and has probably been already reshaped"
                )
                reshaped = True
            else:
                # Reshape
                self.set_data_raw(
                    measurement,
                    self._reshape_frames(
                        self.get_data_raw(measurement),
                        self.nframes,  # number of frames
                        order=self.cfg.files["oscillo"].order,
                    ),
                )
                newshape = self.data_raw[measurement].shape
                logger.debug(f"Reshaped to {newshape}")
                reshaped = True

        if reshaped:
            if "dt_acq" not in self.metadata:
                if not self.is_digital:
                    self.load_metadata()
                else:
                    logger.error(
                        "Oscilloscope data was not loaded, could not get acquisition "
                        "sampling rate. Load the data then re-run reshape_frames()"
                    )
            # Number of time points in a frame
            self.nsamples = self.data_raw[measurement].shape[0]
            # Duration of a frame
            frame_duration = self.nsamples * self.metadata["dt_acq"]
            self.metadata["frame_duration"] = frame_duration  # store it as metadata
            # Build time vector
            measurement_time = np.linspace(
                0, frame_duration - self.metadata["dt_acq"], self.nsamples
            )
            if microseconds:
                # convert to microseconds
                measurement_time *= 1e6
                self.is_us = True
            else:
                self.is_us = False
            # Store
            self.set_data_raw("time_meas", measurement_time)
            self.metadata["toffset_meas"] = self.get_data_raw("time_meas")[0]

        return self

    def get_time_exp(self) -> Self:
        """
        Build the experiment time vector from frame onsets.

        The time point associated to a frame is assumed to be at the middle of the
        duration of a frame. The resulting vector has a shape of (nframes,) and is
        stored in `data_processed[results]`.
        """
        logger.info("Building the experiment time vector...")
        # Checks
        if "frame_duration" not in self.metadata:
            # we don't know how long a frame is, attempt to get through reshape
            self.reshape_frames()
            # won't do anything if oscilloscope data not loaded so we need to recheck
            if "frame_duration" not in self.metadata:
                logger.error(
                    "Oscilloscope data not loaded, can't determine frame duration"
                )
                return self

        # Get parameters
        seq_onsets = self.get_data_raw("frame_onsets")
        seq_duration = self.metadata["frame_duration"]

        # Build the the time vector : attribute a point to the middle of its frame
        # time range
        self.set_data_processed("time_exp", seq_onsets + seq_duration / 2)

        # Add time interval in metadata
        dt_exp = np.mean(np.diff(self.get_data_processed("time_exp")))
        self.metadata["dt_exp"] = dt_exp

        # Verbose
        logger.info(f"Experiment time vector built with dt={dt_exp}s")

        return self

    def average_frame_range(
        self, tstart: None | float = None, tstop: None | float = None, **kwargs
    ) -> Self:
        """
        Average signal in the given frame time range.

        `tstart` and `tstop` are expressed in unit of time (s). If they are not
        provided, they are read from the Config settings ('analysis_window').

        If the averaging is successful :
        1. The corresponding serie NXgroup is created as
            `data_processed[results_{serie_name}{serie_idx}]`,
        2. The analysis window is set as an attribute,
        3. The mean signal amplitude and phase are derived from the I and Q components
            and stored in this serie group.
        """
        # Reset averaging success flag
        self.is_averaged = False
        # Get parameters
        measurements = self.measurements  # list of arrays to average
        if tstart is None:
            if len(self.analysis_window) == 2:
                tstart = self.analysis_window[0]
            else:
                tstart = np.inf
        if tstop is None:
            if len(self.analysis_window) == 2:
                tstop = self.analysis_window[1]
            else:
                tstop = -np.inf
        if not self.is_us:
            # analysis_window should be in microseconds
            toffset = self.metadata["toffset_meas"] * 1e6
        else:
            toffset = self.metadata["toffset_meas"]

        # Check the range
        if tstop < tstart:
            logger.error(f"Invalid time range : {[tstart, tstop]}")
            return self

        # Prepare storage
        self.create_data_serie()

        # Perform average
        for measurement in measurements:
            # Get data
            data = self.get_measurement_data(measurement)
            # Get the actual data name for logging
            meas_name = (
                f"s{measurement}" if self.is_rollmean[measurement] else measurement
            )
            if data is None:
                logger.error(f"{meas_name} not found, not averaging")
                return self

            suffix = "µs" if self.is_us else "s"
            logger.info(
                f"Averaging {measurement} (using {meas_name}) in [{tstart}{suffix}, "
                f"{tstop}{suffix}]..."
            )

            # Determine time interval
            if self.is_decimated:
                # Decimation was applied
                dt = np.mean(np.diff(self.get_data_processed("time_meas")))
                if self.is_us:
                    # Convert to seconds
                    dt *= 1e-6
            else:
                # Original time interval
                dt = self.metadata["dt_acq"]

            # Average & store (keep the original measurement name)
            self.set_data_serie(
                measurement + "_avg",
                self._average_frame_range(
                    data, tstart, tstop, dt, toffset=toffset, **kwargs
                ),
            )
            self.is_averaged = True  # averaging worked
            logger.info(f"{measurement} averaged")

        # Compute phase from I and Q
        if not self.is_averaged:
            # Did not work somehow, most likely messages were logged before
            return self

        # If averaging worked, we can compute amplitude and phase on the average
        logger.info("Computing amplitude...")
        self.compute_amplitude_avg()
        logger.info("Computing phase...")
        self.compute_phase_avg()
        logger.info("Average amplitude and phase computed")

        # Store analysis window
        self.metadata["last_analysis_window"] = [tstart, tstop]
        self.set_attr_serie("analysis_window", [tstart, tstop])
        self.set_attr_serie("analysis_window_unit", "µs")

        # Get time vector
        if not self.get_data_processed("time_exp", checkonly=True):
            self.get_time_exp()
        # Get magnetic field
        if not self.get_data_processed("magfield", checkonly=True):
            self.compute_field().align_field()

        # Prepare placeholders that will be replaced by actual NXlink when the
        # object has a NXroot entry
        phtime = f"!link to:processed/analysis/{self._results_name}/time_exp"
        phmagfield = f"!link to:processed/analysis/{self._results_name}/magfield"

        # Store the link in the 'serie' group
        self.set_data_serie("time_exp", phtime)
        self.set_data_serie("magfield", phmagfield)

        return self

    def compute_amplitude_avg(self) -> Self:
        """
        Compute amplitude of averaged time series.

        In analog mode, amplitude is directly an oscilloscope channel, in digital mode,
        it needs to be derived from I and Q.
        """
        if self.is_digital:
            self._compute_amplitude_avg()
        else:
            # Mark for success
            self.is_averaged = True

        return self

    def compute_phase_avg(self) -> Self:
        """Compute phase from I and Q and correct for pi jumps."""
        # Get names
        iname = self._i_name + "_avg"
        qname = self._q_name + "_avg"

        # Get data
        i = self.get_data_serie(iname)
        q = self.get_data_serie(qname)

        # Compute phase
        res = sp.compute_phase_iq(
            i,
            q,
            unwrap=True,
            period=self.cfg.settings.max_phase_jump * 2 * np.pi,
            axis=0,
        )

        # Store
        self.set_data_serie("phase_avg", res)

        # Mark for success
        self.is_averaged = True

        return self

    def compute_attenuation(self) -> Self:
        """
        Compute attenuation from amplitude average for the current echo.

        It is computed with the `sp.compute_attenuation()` function from the
        averaged amplitude and stored in the current serie group.
        """
        # Get parameters
        if "dt_exp" not in self.metadata:
            self.get_time_exp()
        if not self.is_averaged or not self.get_data_serie(
            "amplitude_avg", checkonly=True
        ):
            logger.warning(
                "Averaging amplitude or phase (or both) failed, cannot compute "
                "attenuation"
            )
            return self

        # Get data
        amp_avg = self.get_data_serie("amplitude_avg")

        # Get baseline
        wbline = self.cfg.settings.range_baseline
        dt = self.metadata["dt_exp"]
        idx_start = int(np.ceil(wbline[0] / dt))
        idx_stop = int(np.ceil(wbline[1] / dt))
        amp0 = amp_avg[idx_start:idx_stop].mean()

        # Compute
        logger.info(f"Computing attenuation for echo index {self.idx_serie}...")
        self.set_data_serie(
            "attenuation",
            sp.compute_attenuation(
                amp_avg,
                amp0,
                self.idx_serie,
                self.cfg.parameters.sample_length,
                mode=self.cfg.parameters.detection_mode,
                scale="linear" if self.is_digital else "log",
                corr=self.cfg.parameters.logamp_slope,
            ),
        )
        max_att = self.get_data_serie("attenuation").max()
        logger.info(f"Done, max. attenuation : {max_att / 100:2.2f}dB/cm")

        return self

    def compute_phase_shift(self) -> Self:
        """
        Compute relative phase shift given the phase average for the current echo.

        It is computed with the `sp.compute_phase_shift()` function from the
        averaged phase and stored in the current serie group.
        """
        # Get parameters
        if "dt_exp" not in self.metadata:
            self.get_time_exp()
        if not self.is_averaged or not self.get_data_serie("phase_avg", checkonly=True):
            logger.warning(
                "Averaging amplitude or phase (or both) failed, cannot compute "
                "phase shift"
            )
            return self

        # Get data
        phi_avg = self.get_data_serie("phase_avg")

        # Get baseline
        wbline = self.cfg.settings.range_baseline
        dt = self.metadata["dt_exp"]
        idx_start = int(np.ceil(wbline[0] / dt))
        idx_stop = int(np.ceil(wbline[1] / dt))
        phi0 = phi_avg[idx_start:idx_stop].mean()

        # Compute
        logger.info(f"Computing phase shift for echo index {self.idx_serie}...")
        self.set_data_serie(
            "phaseshift",
            sp.compute_phase_shift(
                phi_avg,
                phi0,
                self.idx_serie,
                self.cfg.parameters.sample_speed,
                self.cfg.demodulation.f0
                if self.is_digital
                else self.metadata["rf_frequency"],
                self.cfg.parameters.sample_length,
                mode=self.cfg.parameters.detection_mode,
            ),
        )
        max_phi = self.get_data_serie("phaseshift").max()
        logger.info(f"Done, max. relative phase shift : {max_phi * 100:.6f}%")

        return self

    def compute_amplitude(self) -> np.ndarray:
        """
        Compute amplitude from raw or demodulated data (not averaged).

        Either return directly amplitude from 'data_raw' in analog mode, or derive it
        from I and Q in digital mode.
        The returned array has shape (nframes, npoints).
        """
        if self.is_digital:
            amp = sp.compute_amp_iq(
                self.get_data_processed(self._i_name),
                self.get_data_processed(self._q_name),
            )
        else:
            amp = self.get_data_raw("amplitude")

        return amp

    def compute_phase(self) -> np.ndarray:
        """
        Compute phase from raw or demodulated data (not averaged).

        Phase is the arctan of the I and Q components, corrected for pi-jumps.
        The returned array has shape (nframes, npoints).
        """
        if self.is_digital:
            phase = sp.compute_phase_iq(
                self.get_data_processed(self._i_name),
                self.get_data_processed(self._q_name),
            )
        else:
            phase = sp.compute_phase_iq(
                self.get_data_raw(self._i_name), self.get_data_raw(self._q_name)
            )

        return phase

    def find_f0(self) -> Self:
        """
        Find center frequency in reference signal.

        The resulting frequency will be used to build the continuous reference wave used
        for demodulating the signal. It is stored in Config object.

        It uses the `sp.find_f0()` function that relies on RFFT.

        Only the part of the reference signal where there is an actual signal (as found
        by `_find_signal_in_ref()`) will be used to find f0. Optionally, detrending can
        be applied for better results of the FFT. This is controlled with the "detrend"
        parameter in the configuration.
        """
        # Check there is something to do
        if not self.is_digital:
            logger.warning("Can't find center frequency in analog mode")
            return self
        else:
            # for type checking
            assert self.cfg.demodulation is not None

        # Get parameters
        if self._ref_name not in self.data_raw:
            logger.error("Reference signal was not loaded, can't find center frequency")
            return self

        if "ref_on" not in self.metadata:
            # find reference signal onset and offset
            self._find_signal_in_ref()

        tstart = self.metadata["ref_on"]
        tstop = self.metadata["ref_off"]
        nframes = self.cfg.demodulation.fft_nframes
        if nframes > 0:
            # get random frames
            rng = np.random.default_rng(self._seed)
            framesid = rng.integers(0, self.nframes, nframes)
            sig = self.get_data_raw(self._ref_name)[tstart:tstop, framesid]
        else:
            # take all frames
            sig = self.get_data_raw(self._ref_name)[tstart:tstop, :]

        # Optional detrending
        if self.cfg.demodulation.detrend:
            logger.info("Detrending reference signal before FFT...")
            sig = signal.detrend(sig, axis=0)

        logger.info("Finding center frequency in reference signal...")
        self.cfg.demodulation.f0 = sp.find_f0(sig, 1 / self.metadata["dt_acq"]).mean()

        logger.info(f"Found {self.cfg.demodulation.f0 / 1e6:3.6f}MHz.")

        return self

    def demodulate(self, **kwargs) -> Self:
        """
        Digital demodulation : frequency-shifting followed by a low-pass filter.

        A reference signal and a measurement signal are required, as well as a center
        frequency.

        The reference signal is fitted to extract the phase (when there is signal, see
        the `_find_signal_in_ref()` method) and build a continuous, mono-frequency
        signal. The measurement signal is frequency-shifted (multiplied by cos(phi_ref)
        and sin(phi_ref)), then a low-pass filter is applied. Optionally, the frequency-
        shifted signal can be decimated before applying the filter. The filter
        properties are set in the configuration.

        The demodulation is done in chunks to not saturate memory (chunks size is set in
        the configuration).

        The I and Q components are stored in the `data_processed[results]`. If
        decimation is used, the corresponding time vector is updated accordingly and
        stored there as well.

        Uses the `sp.demodulate_chunks()` function.
        """
        if not self.is_digital:
            logger.warning("Can't perform demodulation in analog mode")
            return self
        else:
            # for type checking
            assert self.cfg.demodulation is not None

        # Check data is loaded
        if "time_meas" not in self.data_raw:
            logger.error("Data was not loaded, not demodulating")
            return self
        else:
            if self.is_us:
                tmeas = self.get_data_raw("time_meas") * 1e-6  # convert back to seconds
            else:
                tmeas = self.get_data_raw("time_meas")
        # Check reference was scanned for signal
        if "ref_on" not in self.metadata:
            self._find_signal_in_ref()
        # Check what rf frequency will be used
        if self.cfg.demodulation.f0 > 0:
            # f0 set in the configuration file
            logger.info("f0 set in the configuration")
        elif self.cfg.demodulation.auto_find_f0:
            # find f0 automatically
            logger.info("f0 not set, finding automatically...")
            self.find_f0()
        else:
            # use the frequency read in the metadata
            logger.info("f0 not set, using RF frequency read in metadata")
            self.cfg.demodulation.f0 = self.metadata["rf_frequency"]

        # Get parameters for reference fitting range
        istart = self.metadata["ref_on"]
        istop = self.metadata["ref_off"]

        logger.info(f"Demodulation at {self.cfg.demodulation.f0 / 1e6:3.6f}MHz...")
        # Free some space
        self.remove_data_processed(self._i_name)
        self.remove_data_processed(self._q_name)
        in_phase, out_phase = sp.demodulate_chunks(
            tmeas[istart:istop],
            self.get_data_raw(self._ref_name)[istart:istop, :],
            tmeas,
            self.get_data_raw(self._sig_name),
            f0=self.cfg.demodulation.f0,
            filter_order=self.cfg.demodulation.filter_order,
            filter_fc=self.cfg.demodulation.filter_fc,
            decimate_factor=self.cfg.demodulation.decimate_factor,
            chunksize=self.cfg.demodulation.chunksize,
            **kwargs,
        )
        logger.info("Signal demodulated")

        # Store
        self.set_data_processed(self._i_name, in_phase)  # I
        self.set_data_processed(self._q_name, out_phase)  # Q

        # In case of decimation, the time vector needs to be updated
        if self.cfg.demodulation.decimate_factor > 1:
            logger.info("Decimation was used, updating time vector...")
            # Get new number of points and create new time vector
            self._update_time_subsampled(in_phase.shape[0])
            # Update flag
            self.is_decimated = True
            logger.info(f"Done, I and Q now have {in_phase.shape[0]} samples")
        else:
            # Remove decimated time vector
            self.remove_data_processed("time_meas")
            self.is_decimated = False

        # Remove rolling-averaged data and update flags
        self.is_rollmean.update({m: False for m in self.measurements})
        for meas in self.measurements:
            self.remove_data_processed(f"s{meas}")
        # Update time vector that was maybe subsampled because of rolling average
        self.get_time_exp()
        self.align_field()

        return self

    def get_csv_filename(self) -> str:
        """
        Generate output CSV file name for the current echo.

        The file is placed in the data directory set in the configuration. Its name
        contains the analysis time window and the echo number.
        """
        fout = os.path.join(self.cfg.data_directory, self.cfg.expid + "-results")
        echo_index = self.idx_serie
        if "last_analysis_window" not in self.metadata:
            logger.error("Analysis window was never set")
            tstart = -1
            tstop = -1
        tstart = self.metadata["last_analysis_window"][0]
        tstop = self.metadata["last_analysis_window"][1]
        return f"{fout}_{tstart:.3f}-{tstop:.3f}_echo{echo_index}.csv"

    def get_measurement_data(self, meas: str) -> np.ndarray | None:
        """
        Get the dataset corresponding to the measurement `meas`.

        Determine where to look up the data, depending if we need the raw, smoothed or
        demodulated data.

        If the data cannot be found, it returns None.
        """
        if self.is_rollmean[meas]:
            # Smoothed trace (rolling average applied)
            if self.get_data_processed(f"s{meas}", checkonly=True):
                return self.get_data_processed(f"s{meas}")
            else:
                return None
        if self.is_digital:
            # Demodulated trace
            if self.get_data_processed(meas, checkonly=True):
                return self.get_data_processed(meas)
            else:
                return None
        else:
            # Raw trace
            if self.get_data_raw(meas, checkonly=True):
                return self.get_data_raw(meas)
            else:
                return None

    def _init_flags(self) -> None:
        """Set up flags."""
        # Tektronix WFM files or LabVIEW binary file
        self.is_wfm = False
        # Measurement time vector is in microseconds
        self.is_us = False
        # Averaged values were computed or updated
        self.is_averaged = False
        # Where the corresponding time vector is stored ('data_processed' if decimation
        # is used)
        self.is_decimated = False

        # Determine analog or digital mode, if needed
        if (self.is_digital is None) and self.is_config_file:
            self.is_digital = self._check_digital()

        # Moving mean was performed, if it was, use the smoothed data in data_processed
        self.is_rollmean = {m: False for m in self.measurements}

    def _guess_flags(self) -> None:
        """Attempt to guess flags from data after loading a NeXus file."""
        self.is_wfm = None  # no files so irrelevant
        self.is_us = self._check_measurement_time_us()
        self.is_averaged = self._check_averaged()

        if self.is_digital is None:
            # It was not set by the user when initializing the Processor object
            self.is_digital = self._check_digital()

        if self.is_digital:
            self._init_digital()  # fills is_meas_processed
            self._sig_name = self._get_signal_name()

        self.is_decimated = self._check_time_subsampled()
        self.is_rollmean = self._check_rollmean()

    def _get_signal_name(self) -> str:
        """Determine signal name from configuration file measurements."""
        signames = [*{*self.cfg.measurements.keys()} - {self._ref_name}]
        if len(signames) != 1:
            raise ValueError(
                "Config file : Measurements should have 2 and only 2 entries."
            )
        return signames[0]

    @staticmethod
    def _load_metadata(
        filename: str | Path,
        header_map: dict[str, int],
        conversion_map: dict[str, bool],
    ) -> dict[str, float | str]:
        """
        Read metadata from the header of the text file with the frames onsets.

        `header_map` is a dict mapping a metadata to a line number in the file
        (0-based).

        `conversion_map` is a dict specifying if those metadata should be
        converted to numerical values.

        Parameters
        ----------
        filename : str | Path
            Full path to the text file with metadata.
        header_map : dict
            Maps a metadata name to a line number in the file (0-based).
        conversion_map : dict
            Maps a metadata name to a booelan, requiring if the metadata should be
            converted to a numerical value or kept as-is.

        Returns
        -------
        metadata : dict
            {metadata name : value}
        """
        # Determine how many lines-long the header is
        nlines_header = max(header_map.values()) + 1

        # Check file exists
        if not os.path.isfile(filename):
            logger.warning(
                f"{os.path.basename(filename)} : metadata file does not exist"
            )
            return {}

        # Read first lines
        with open(filename) as fid:
            header = [next(fid) for _ in range(nlines_header)]

        # Collect
        metadata = {key: header[idx] for key, idx in header_map.items()}

        # Convert numerical values
        metadata = {
            key: float(value) if conversion_map[key] else value.strip("\n")
            for key, value in metadata.items()
        }

        return metadata

    @staticmethod
    def _load_frame_onsets(
        filename: str | Path, nlines_header: int, delimiter: str
    ) -> np.ndarray:
        """
        Read frame onsets in the reference time text file.

        Parameters
        ----------
        filename : str | Path
            Full path to the reference time text file.
        nlines_header : int
            Number of lines to ignore to skip the header.
        delimiter : str
            Delimiter in the text file.

        Returns
        -------
        frames_onsets : np.ndarray
            Vector with frames onsets.
        """
        frame_onsets = np.loadtxt(
            filename,
            skiprows=nlines_header,
            delimiter=delimiter,
            usecols=0,
            dtype=float,
        )
        return frame_onsets

    def _load_oscillo_bin(
        self,
        filename: str | Path,
        nchannels: int,
        precision: int,
        endian: Literal["<", ">"] = "<",
        order: Literal["F", "C"] = "F",
    ) -> np.ndarray:
        """
        Load oscilloscope binary file.

        Loaded data is reshaped and transposed so that time series corresponding to one
        channel correspond to one row (so that it can be easily unpacked). Note that the
        time serie is 1D : all frames are stored one after the other and need to be
        reshaped.

        Parameters
        ----------
        filename : str | Path
            Full path to binary file.
        nchannels : int
            Number of channels.
        precision : int
            Byte precision.
        endian : {"<", ">"}, optional
            "<" for little endian, ">" for big endian. Default is "<".
        order : {"F", "C"}, optional
            Array order, "F" for Fortran, "C" for C. Default is "F".

        Returns
        -------
        oscillo_data np.ndarray
            Shape nchannels * nsamples.

        Raises
        ------
        ValueError
            If it can't be reshaped to the expected shape, raises an error.
        """
        if not isinstance(filename, Path):
            filename = Path(filename)

        # Load data
        data = self.load_bin(filename, precision, endian)

        # Reshape data as an array with a column per channel
        filesize = filename.stat().st_size  # get total size in bytes

        # Normally there are filesize bytes in total in the file, precision bytes per
        # float, so 12 bytes per sample (4 bytes * 3 channels), hence
        # filesize / (nchannels * precision) columns and nchannels lines
        # If the data is correct, it should be possible to reshape like this, as all
        # channels have the same number of time points.
        ncolumns = int(filesize / (nchannels * precision))
        try:
            data = data.reshape((ncolumns, nchannels), order=order).astype(float)
        except ValueError:
            raise ValueError("Missing samples in the oscilloscope binary file.")

        return data.T

    @staticmethod
    def _reshape_frames(
        oscillo_data: np.ndarray,
        nframes: int,
        order: Literal["F", "C"] = "F",
    ) -> np.ndarray:
        """
        Reshape a 1D array into a 2D array.

        Time series will end up on columns.

        Parameters
        ----------
        oscillo_data : np.ndarray
            1D array with time gaps.
        nframes : int
            Number of frames, the resulting array will have `nframes` columns.
        order : {"F", "C"}, optional
            Order, "F" for Fortran, "C" for C. Default is "F".

        Returns
        -------
        oscillo_data : np.ndarray
            Reshaped array.
        """
        nsamples_per_seq = int(oscillo_data.shape[0] // nframes)

        return oscillo_data.reshape((nsamples_per_seq, nframes), order=order)

    @staticmethod
    def _average_frame_range(
        frame: np.ndarray,
        tstart: float,
        tstop: float,
        dt: float,
        toffset: float = 0,
        tunit: str = "us",
    ) -> np.ndarray:
        """
        Average the signal in the `frame` 2D array in the given time window.

        The time window is defined by `tstart` and `tstop` expressed in the frame time
        base (`dt`).

        Parameters
        ----------
        frame : np.ndarray
            Array with time series on columns.
        tstart, tstop : float
            Defines the time window in which the average is computed.
        dt : float
            Time interval between two samples, in seconds.
        toffset : float, optional
            Offset in time, if the measurement time vector does not start at 0, default
            is 0.
        tunit : {"us", "ms", "s"} str, optional
            Units in which tstart and tstop are given, default is "us" (microseconds).

        Returns
        -------
        meas_avg : np.ndarray
            Measurement averaged in-frame, so that the resulting 1D vector is aligned on
            the `time_exp` vector.
        """
        match tunit:
            case "us":
                mult = 1e-6
            case "ms":
                mult = 1e-3
            case "s":
                mult = 1
            case _:
                logger.warning(
                    f"{tunit} is not recognized as a valid unit, assuming it is 'us'"
                )
                mult = 1e-6

        idx_start = int(np.ceil((tstart - toffset) * mult / dt))
        idx_stop = int(np.ceil((tstop - toffset) * mult / dt))

        return frame[idx_start:idx_stop, :].mean(axis=0)

    def _compute_amplitude_avg(self) -> None:
        """
        Compute amplitude from averaged I and Q.

        They are read from `data_processed[results][results_{serie_name}{idx_name}]`,
        and the results is stored as "amplitude_avg".
        """
        # Get names
        iname = self._i_name + "_avg"
        qname = self._q_name + "_avg"

        # Get data
        if not self.get_data_serie(iname, checkonly=True) or not self.get_data_serie(
            qname, checkonly=True
        ):
            logger.warning(
                f"'{iname}' or '{qname} not found, was the signal demodulated before "
                "averaging ?"
            )
            self.is_averaged = False
            return

        i = self.get_data_serie(iname)
        q = self.get_data_serie(qname)

        # Compute averaged amplitude
        self.set_data_serie("amplitude_avg", sp.compute_amp_iq(i, q))

        # Mark for success
        self.is_averaged = True

    def _find_signal_in_ref(self) -> Self:
        """
        Find signal in reference time series.

        Wrap the `sp.find_signal()` function, fetching the parameters from the
        configuration.
        """
        if not self.is_digital:
            logger.warning("Can't find reference signal in analog mode")
            return self
        else:
            # for type checking
            assert self.cfg.demodulation is not None

        # Get parameters
        nframes = self.cfg.demodulation.findsig_nframes
        if nframes > 0:
            rng = np.random.default_rng(self._seed)
            framesid = rng.integers(0, self.nframes, nframes)
            sig = self.get_data_raw(self._ref_name)[:, framesid]
        else:
            sig = self.get_data_raw(self._ref_name)
        std_factor = self.cfg.demodulation.findsig_nstd
        before = self.cfg.demodulation.findsig_extend

        # Detect signal
        logger.info("Detecting signal in reference trace...")
        start, stop = sp.find_signal(sig, std_factor, before=before, after=before)

        # Store
        self.metadata["ref_on"] = start
        self.metadata["ref_off"] = stop

        logger.info(f"Reference signal cropped between indices [{start, stop}]")

        return self

    def _update_time_subsampled(self, new_npoints: int) -> None:
        """
        Update time vector if subsampling was used.

        If decimation was used during demodulation, the corresponding time vector needs
        to be updated and placed in 'data_processed'.
        """
        self.set_data_processed(
            "time_meas",
            np.linspace(
                self.get_data_raw("time_meas")[0],
                self.get_data_raw("time_meas")[-1],
                new_npoints,
            ),
        )

    def _check_frames_2d(self, meas_name: str) -> bool:
        """
        Check if the array corresponding to `meas_name` is 1 or 2D.

        Returns False also if the dataset can't be found.

        Parameters
        ----------
        meas_name : str
            Name of some measurement in `data_raw`.
        where : {"raw", "processed"}, optional
            Whether to look up the data in 'data_raw' (default) or in 'data_processed'.

        Returns
        -------
        res : bool
            Returns true if the array is 2D, or if `reshape_frame()` was called.
        """
        data = self.get_measurement_data(meas_name)

        if data is None or data.ndim != 2:
            return False
        else:
            return True

    def _check_measurement_time_us(self) -> bool:
        """
        Check if the measurement time is in microseconds.

        Read the `time_meas` dataset attribute "units", if it does not exist, look at
        the order of magnitude of values in the vector and compare to the
        `EXP_TIMESCALE` global variable. If the maximum is 1000 times higher, it is
        assumed the vector is in microseconds.

        If the measurement time vector is not found, issue a warning and return True as
        it should be the default.
        """
        if "time_meas" in self.data_raw:
            if "units" in self.data_raw["time_meas"].attrs:
                # Check attribute
                if self.data_raw["time_meas"].attrs["units"] in ("us", "µs"):
                    return True
                else:
                    return False
            else:
                if self.data_raw["time_meas"].max() > 1e3 * FRAME_TIMESCALE:
                    # Check numbers scale
                    return True
        logger.warning("Could not determine 'time_meas' units, guessing µs (default).")
        return True

    def _check_averaged(self) -> bool:
        """
        Check if data was averaged.

        This is done by checking if any dataset name in `data_processed[results]` ends
        with "_avg" as written by the `average_frame_range()` method.
        """
        pattern = "_avg"
        for item in self.data_processed.values():
            if isinstance(item, nx.NXdata):
                for key in item:
                    if key.endswith(pattern):
                        return True

        return False

    def _check_digital(self) -> bool:
        """
        Check if the input should be treated as analog or digital.

        This is done by checking if the configuration has a "demodulation" section. Note
        that even if this section is empty, True will be returned.
        """
        # Check if the Config object has a non-None 'demodulation' attribute
        res = getattr(self.cfg, "demodulation", None)
        if res is None:
            return False
        else:
            return True

    def _check_rollmean(self) -> dict[str, bool]:
        """
        Check if rolling average was applied for each measurement.

        This is done by checking if measurements datasets are found in
        `data_processed[results]` with a name with a leading "s", as written by the
        `rolling_average()` method.
        """
        is_rollmean = dict()
        for meas in self.measurements:
            if self.get_data_processed(f"s{meas}", checkonly=True):
                is_rollmean[meas] = True
            else:
                is_rollmean[meas] = False
        return is_rollmean

    def _check_time_subsampled(self) -> bool:
        """
        Check if measurement time vector was subsampled.

        Happens when decimation is used during demodulation. This is done by checking
        if there is a "time_meas" dataset in `data_processed[results]` as this indicates
        some processing was applied to this vector.
        """
        return self.get_data_processed("time_meas", checkonly=True)



"""The Configuration section of the GUI."""

from PyQt6.QtCore import pyqtSignal, pyqtSlot
from pyqtgraph.parametertree import Parameter

from pymagnetos.core.gui.widgets import BaseConfigurationWidget

from ._param_content import ParamContent


class ConfigurationWidget(BaseConfigurationWidget):
    """
    Configuration Parameter Tree for TDO experiments.

    Signals
    -------
    sig_syncroi_changed : emitted when the "Sync ROI" checkbox is changed.
    """

    sig_syncroi_changed = pyqtSignal()

    def __init__(self, param_content: type[ParamContent]):
        super().__init__(param_content)

        self._freq2time = 1.0

        self.add_extra_parameters()

    def add_extra_parameters(self):
        """Add extra parameters used only in the GUI."""
        # Sync ROIs parameter
        self.syncroi_parameter = Parameter.create(
            name="syncroi",
            type="bool",
            value=False,
            title="Sync. fit and FFT field-window",
        )
        self.host_parameters.addChild(self.syncroi_parameter)
        self.syncroi_parameter.sigValueChanged.connect(self.sig_syncroi_changed.emit)

        # Show nperseg parameter in µs
        nperseg = Parameter.create(
            name="spectro_time_window",
            type="float",
            readonly=True,
            value=1024 / 100e6,
            suffix="s",
            siPrefix=True,
            title="Spectro: time window (n/persegment)",
        )
        self.host_parameters.addChild(nperseg)
        self.settings_parameters.child("spectro_nperseg").sigValueChanged.connect(
            self.update_spectro_time_window
        )

    @pyqtSlot()
    def update_spectro_time_window(self):
        """Update the spectro. nperseg setting expressed in time."""
        self.host_parameters["spectro_time_window"] = (
            self.settings_parameters["spectro_nperseg"] / self._freq2time
        )



"""A logging handler to print log stream in a text window."""

import logging

from PyQt6 import QtWidgets
from PyQt6.QtCore import QObject, pyqtSignal


class TextLoggerWidget(logging.Handler, QObject):
    """
    Logger handler to print log in a QTextEdit widget.

    pyqtSignals
    -------
    sig_append_text : internal, emits the log message to be printed in the text box.
    """

    sig_append_text = pyqtSignal(str)
    flushOnClose = True

    def __init__(self, parent):
        super().__init__()
        QObject.__init__(self)

        self.widget = QtWidgets.QPlainTextEdit(parent)
        self.widget.setReadOnly(True)

        self.sig_append_text.connect(self.widget.appendPlainText)

    def emit(self, record: logging.LogRecord) -> None:
        msg = self.format(record)
        self.sig_append_text.emit(msg)



"""Parameters for the `pytdo` configuration."""

import numpy as np

from pymagnetos.core.gui.widgets import BaseParamContent


class ParamContent(BaseParamContent):
    PARAMS_TO_PARSE = ["poly_window", "fft_window"]
    children_parameters = [
        dict(
            name="pickup_surface",
            type="float",
            limits=[0, np.inf],
            value=None,
            suffix="m²",
            siPrefix=False,
            readonly=True,
            title="Pickup surface",
            tip="Surface of the pickup coil in m²",
        )
    ]

    children_settings = [
        dict(
            name="max_time",
            type="float",
            limits=[0, np.inf],
            step=0.1,
            value=1,
            suffix="s",
            siPrefix=True,
            title="Max. time",
            tip="Stop analysis after this duration (in seconds)",
        ),
        dict(
            name="spectro_nperseg",
            type="int",
            limits=[0, np.inf],
            value=1024,
            step=512,
            title="Spectro: n/segment",
            tip="Number of samples per segment for the spectrogram",
        ),
        dict(
            name="spectro_win_size",
            type="int",
            limits=[2, np.inf],
            value=2048,
            step=1024,
            title="Spectro: FFT padding size",
            tip="Padding size for the FFT in each segment for the spectrogram",
        ),
        dict(
            name="spectro_noverlap",
            type="int",
            limits=[-1, np.inf],
            value=-1,
            step=512,
            title="Spectro: n overlap (-1 : n/2)",
        ),
        dict(
            name="barycenters_fwindow",
            type="float",
            limits=[0, np.inf],
            step=1e3,
            value=1,
            suffix="Hz",
            siPrefix=True,
            title="Barycenters: frequency window",
        ),
        dict(
            name="barycenters_fast",
            type="bool",
            limits=[True, False],
            value=True,
            title="Barycenters: fast algo.",
        ),
        dict(
            name="time_offset",
            type="float",
            limits=[-np.inf, np.inf],
            suffix="s",
            siPrefix=True,
            value=0e-6,
            step=0.5e-6,
            title="TDO/Pickup time offset",
        ),
        dict(name="poly_window", type="str", value="", title="Fit: field window"),
        dict(
            name="poly_deg",
            type="int",
            value=3,
            step=1,
            limits=[1, 10],
            title="Fit: degree",
        ),
        dict(
            name="npoints_interp_inverse",
            type="int",
            value=100000,
            step=5e4,
            limits=[1000, np.inf],
            title="n points interp. 1/B",
        ),
        dict(
            name="fft_window",
            type="str",
            value="",
            title="FFT: field window",
        ),
        dict(
            name="fft_pad_mult",
            type="int",
            value=100,
            step=100,
            title="FFT: padding size mult.",
        ),
        dict(
            name="max_bfreq",
            type="float",
            value=70000,
            step=1e4,
            limits=[1, np.inf],
            suffix="T",
            siPrefix=True,
            title="FFT : Max. B-frequency",
        ),
        dict(
            name="offset",
            type="float",
            value=1000,
            step=500,
            limits=[0, np.inf],
            title="Display: curve offset",
        ),
    ]
